import sys
import types

from .s3 import MetaflowS3Exception, S3

# Import any additional datatools defined by a Metaflow custom package
try:
    import metaflow_custom.datatools as extension_module
except ImportError as e:
    ver = sys.version_info[0] * 10 + sys.version_info[1]
    if ver >= 36:
        # e.name is set to the name of the package that fails to load
        # so don't error ONLY IF the error is importing this module (but do
        # error if there is a transitive import error)
        if not (isinstance(e, ModuleNotFoundError) and \
                e.name in ['metaflow_custom', 'metaflow_custom.datatools']):
            print(
                "Cannot load metaflow_custom exceptions -- "
                "if you want to ignore, uninstall metaflow_custom package")
            raise
else:
    # We load into globals whatever we have in extension_module
    # We specifically exclude any modules that may be included (like sys, os, etc)
    # *except* for ones that are part of metaflow_custom (basically providing
    # an aliasing mechanism)
    lazy_load_custom_modules = {}
    addl_modules = extension_module.__dict__.get('__mf_promote_submodules__')
    if addl_modules:
        # We make an alias for these modules which the metaflow_custom author
        # wants to expose but that may not be loaded yet
        lazy_load_custom_modules = {
            'metaflow.datatools.%s' % k: 'metaflow_custom.datatools.%s' % k
            for k in addl_modules}
    for n, o in extension_module.__dict__.items():
        if not n.startswith('__') and not isinstance(o, types.ModuleType):
            globals()[n] = o
        elif isinstance(o, types.ModuleType) and o.__package__ and \
                o.__package__.startswith('metaflow_custom'):
            lazy_load_custom_modules['metaflow.datatools.%s' % n] = o
    if lazy_load_custom_modules:
        from metaflow import _LazyLoader
        sys.meta_path = [_LazyLoader(lazy_load_custom_modules)] + sys.meta_path
finally:
    # Erase all temporary names to avoid leaking things
    for _n in ['ver', 'n', 'o', 'e', 'lazy_load_custom_modules',
               'extension_module', '_LazyLoader', 'addl_modules']:
        try:
            del globals()[_n]
        except KeyError:
            pass
    del globals()['_n']



import subprocess
from .exception import ExternalCommandFailed

from metaflow.util import to_bytes

def cmd(cmdline, input, output):
    for path, data in input.items():
        with open(path, 'wb') as f:
            f.write(to_bytes(data))

    if subprocess.call(cmdline, shell=True):
        raise ExternalCommandFailed("Command '%s' returned a non-zero "
                                    "exit code." % cmdline)

    out = []
    for path in output:
        with open(path, 'rb') as f:
            out.append(f.read())

    if len(out) == 1:
        return out[0]
    else:
        return out



import time
import requests
import json

from threading import Thread
from metaflow.sidecar_messages import MessageTypes, Message
from metaflow.metaflow_config import METADATA_SERVICE_HEADERS
from metaflow.exception import MetaflowException

HB_URL_KEY = 'hb_url'


class HeartBeatException(MetaflowException):
    headline = 'Metaflow heart beat error'

    def __init__(self, msg):
        super(HeartBeatException, self).__init__(msg)

class MetadataHeartBeat(object):

    def __init__(self):
        self.headers = METADATA_SERVICE_HEADERS
        self.req_thread = Thread(target=self.ping)
        self.req_thread.daemon = True
        self.default_frequency_secs = 10
        self.hb_url = None

    def process_message(self, msg):
        # type: (Message) -> None
        if msg.msg_type == MessageTypes.SHUTDOWN:
            # todo shutdown doesnt do anything yet? should it still be called
            self.shutdown()
        if (not self.req_thread.is_alive()) and \
                msg.msg_type == MessageTypes.LOG_EVENT:
            # set post url
            self.hb_url = msg.payload[HB_URL_KEY]
            # start thread
            self.req_thread.start()

    def ping(self):
        retry_counter = 0
        while True:
            try:
                frequency_secs = self.heartbeat()

                if frequency_secs is None or frequency_secs <= 0:
                    frequency_secs = self.default_frequency_secs

                time.sleep(frequency_secs)
                retry_counter = 0
            except HeartBeatException as e:
                retry_counter = retry_counter + 1
                time.sleep(4**retry_counter)

    def heartbeat(self):
        if self.hb_url is not None:
            response = \
                requests.post(url=self.hb_url, data='{}', headers=self.headers)
            # Unfortunately, response.json() returns a string that we need
            # to cast to json; however when the request encounters an error
            # the return type is a json blob :/
            if response.status_code == 200:
                return json.loads(response.json()).get('wait_time_in_seconds')
            else:
                raise HeartBeatException('HeartBeat request (%s) failed'
                                         ' (code %s): %s' % 
                                            (self.hb_url, response.status_code,
                                            response.text))
        return None

    def shutdown(self):
        # attempts sending one last heartbeat
        self.heartbeat()


class ByteStream(object):
    """Basic interface that reads and writes bytes"""

    def read(self, count, timeout=None):
        """
        Reads exactly count bytes from the stream. This call is blocking until count bytes
        are read or an error happens

        This call returns a byte array or EOFError if there was a problem
        reading.

        Parameters
        ----------
        count : int
            Exact number of characters to read

        Returns
        -------
        bytes
            Content read from the stream

        Raises
        ------
        EOFError
            Any issue with reading will be raised as a EOFError
        """
        raise NotImplementedError

    def write(self, data):
        """
        Writes all the data to the stream

        This call is blocking until all data is written. EOFError will be
        raised if there is a problem writing to the stream

        Parameters
        ----------
        data : bytes
            Data to write out

        Raises
        ------
        EOFError
            Any issue with writing will be raised as a EOFError
        """
        raise NotImplementedError

    def close(self):
        """
        Closes the stream releasing all system resources

        Once closed, the stream cannot be re-opened or re-used. If a
        stream is already closed, this operation will have no effect
        """
        raise NotImplementedError()

    @property
    def is_closed(self):
        """
        Returns True if the stream is closed or False otherwise

        Returns
        -------
        bool
            True if closed or False otherwise
        """
        raise NotImplementedError()

    def fileno(self):
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()



CONTROL_TASK_TAG = 'control_task'
UBF_CONTROL = 'ubf_control'
UBF_TASK = 'ubf_task'

class UnboundedForeachInput(object):
    """
    Plugins that wish to support `UnboundedForeach` need their special
    input(s) subclass this class.
    This is used by the runtime to detect the difference between bounded
    and unbounded foreach, based on the variable passed to `foreach`.
    """



from metaflow.plugins.env_escape.override_decorators import (
    local_override,
    local_getattr_override,
    local_setattr_override,
    remote_override,
    remote_getattr_override,
    remote_setattr_override,
    local_exception,
    remote_exception_serialize,
)


@local_override({"test_lib.TestClass1": "print_value"})
def local_print_value(stub, func):
    print("Encoding before sending to server")
    v = func()
    print("Adding 5")
    return v + 5


@remote_override({"test_lib.TestClass1": "print_value"})
def remote_print_value(obj, func):
    print("Decoding from client")
    v = func()
    print("Encoding for client")
    return v

@local_getattr_override({"test_lib.TestClass1": "override_value"})
def local_get_value2(stub, name, func):
    print("In local getattr override for %s" % name)
    r = func()
    print("In local getattr override, got %s" % r)
    return r


@local_setattr_override({"test_lib.TestClass1": "override_value"})
def local_set_value2(stub, name, func, v):
    print("In local setattr override for %s" % name)
    r = func(v)
    print("In local setattr override, got %s" % r)
    return r


@remote_getattr_override({"test_lib.TestClass1": "override_value"})
def remote_get_value2(obj, name):
    print("In remote getattr override for %s" % name)
    r = getattr(obj, name)
    print("In remote getattr override, got %s" % r)
    return r


@remote_setattr_override({"test_lib.TestClass1": "override_value"})
def remote_set_value2(obj, name, v):
    print("In remote setattr override for %s" % name)
    r = setattr(obj, name, v)
    print("In remote setattr override, got %s" % r)
    return r


@local_override({"test_lib.TestClass1": "unsupported_method"})
def unsupported_method(stub, func, *args, **kwargs):
    return NotImplementedError("Just because")


@local_override({"test_lib.package.TestClass3": "thirdfunction"})
def iamthelocalthird(stub, func, val):
    print("Locally the Third")
    v = func(val)
    return v


@remote_override({"test_lib.package.TestClass3": "thirdfunction"})
def iamtheremotethird(obj, func, val):
    print("Remotely the Third")
    v = func(val)
    return v


@local_exception("test_lib.SomeException")
class SomeException:
    def __str__(self):
        parent_val = super(self.__realclass__, self).__str__()
        return parent_val + " In SomeException str override: %s" % self.user_value

    def _deserialize_user(self, json_obj):
        self.user_value = json_obj

@remote_exception_serialize("test_lib.SomeException")
def some_exception_serialize(ex):
    return 42



from __future__ import print_function
import random
import time
import sys
import os

from metaflow.exception import MetaflowException
from metaflow.metaflow_config import S3_ENDPOINT_URL, S3_VERIFY_CERTIFICATE

S3_NUM_RETRIES = 7

TEST_S3_RETRY = 'TEST_S3_RETRY' in os.environ

def get_s3_client():
    from metaflow.plugins.aws.aws_client import get_aws_client
    return get_aws_client(
        's3',
        with_error=True,
        params={'endpoint_url': S3_ENDPOINT_URL, 'verify': S3_VERIFY_CERTIFICATE })

# decorator to retry functions that access S3
def aws_retry(f):
    def retry_wrapper(self, *args, **kwargs):
        last_exc = None
        for i in range(S3_NUM_RETRIES):
            try:
                ret = f(self, *args, **kwargs)
                if TEST_S3_RETRY and i == 0:
                    raise Exception("TEST_S3_RETRY env var set. "
                                    "Pretending that an S3 op failed. "
                                    "This is not a real failure.")
                else:
                    return ret
            except MetaflowException as ex:
                # MetaflowExceptions are not related to AWS, don't retry
                raise
            except Exception as ex:
                try:
                    function_name = f.func_name
                except AttributeError:
                    function_name = f.__name__
                sys.stderr.write("S3 datastore operation %s failed (%s). "
                                 "Retrying %d more times..\n"
                                 % (function_name, ex, S3_NUM_RETRIES - i))
                self.reset_client(hard_reset=True)
                last_exc = ex
                # exponential backoff for real failures
                if not (TEST_S3_RETRY and i == 0):
                    time.sleep(2**i + random.randint(0, 5))
        raise last_exc
    return retry_wrapper


import json
import os
import sys
import time
import shutil
import random
import subprocess
from io import RawIOBase, BytesIO, BufferedIOBase
from itertools import chain, starmap
from tempfile import mkdtemp, NamedTemporaryFile

from .. import FlowSpec
from ..current import current
from ..metaflow_config import DATATOOLS_S3ROOT
from ..util import namedtuple_with_defaults,\
                   is_stringish,\
                   to_bytes,\
                   to_unicode,\
                   to_fileobj,\
                   url_quote,\
                   url_unquote
from ..exception import MetaflowException
from ..debug import debug

try:
    # python2
    from urlparse import urlparse
except:
    # python3
    from urllib.parse import urlparse

from metaflow.datastore.util.s3util import get_s3_client

try:
    import boto3
    boto_found = True
except:
    boto_found = False

def ensure_unicode(x):
    return None if x is None else to_unicode(x)

S3GetObject = namedtuple_with_defaults(
    'S3GetObject', 'key offset length')

S3PutObject = namedtuple_with_defaults(
    'S3PutObject', 'key value path content_type metadata',
    defaults=(None, None, None, None))

RangeInfo = namedtuple_with_defaults(
    'RangeInfo', 'total_size request_offset request_length',
    defaults=(0, -1))

NUM_S3OP_RETRIES = 7

class MetaflowS3InvalidObject(MetaflowException):
    headline = 'Not a string-like object'

class MetaflowS3URLException(MetaflowException):
    headline = 'Invalid address'

class MetaflowS3Exception(MetaflowException):
    headline = 'S3 access failed'

class MetaflowS3NotFound(MetaflowException):
    headline = 'S3 object not found'

class MetaflowS3AccessDenied(MetaflowException):
    headline = 'S3 access denied'


class S3Object(object):
    """
    This object represents a path or an object in S3,
    with an optional local copy.

    Get or list calls return one or more of S3Objects.
    """

    def __init__(
                 self, prefix, url, path,
                 size=None, content_type=None, metadata=None, range_info=None):

        # all fields of S3Object should return a unicode object
        prefix, url, path = map(ensure_unicode, (prefix, url, path))

        self._size = size
        self._url = url
        self._path = path
        self._key = None
        self._content_type = content_type

        self._metadata = None
        if metadata is not None and 'metaflow-user-attributes' in metadata:
            self._metadata = json.loads(metadata['metaflow-user-attributes'])

        if range_info and (range_info.request_length is None or\
                range_info.request_length < 0):
            self._range_info = RangeInfo(
                range_info.total_size,
                range_info.request_offset, range_info.total_size)
        else:
            self._range_info = range_info

        if path:
            self._size = os.stat(self._path).st_size

        if prefix is None or prefix == url:
            self._key = url
            self._prefix = None
        else:
            self._key = url[len(prefix.rstrip('/')) + 1:].rstrip('/')
            self._prefix = prefix

    @property
    def exists(self):
        """
        Does this key correspond to an object in S3?
        """
        return self._size is not None

    @property
    def downloaded(self):
        """
        Has this object been downloaded?
        """
        return bool(self._path)

    @property
    def url(self):
        """
        S3 location of the object
        """
        return self._url

    @property
    def prefix(self):
        """
        Prefix requested that matches the object.
        """
        return self._prefix

    @property
    def key(self):
        """
        Key corresponds to the key given to the get call that produced
        this object. This may be a full S3 URL or a suffix based on what
        was requested.
        """
        return self._key

    @property
    def path(self):
        """
        Path to the local file corresponding to the object downloaded.
        This file gets deleted automatically when a S3 scope exits.

        Returns None if this S3Object has not been downloaded.
        """
        return self._path

    @property
    def blob(self):
        """
        Contents of the object as a byte string.

        Returns None if this S3Object has not been downloaded.
        """
        if self._path:
            with open(self._path, 'rb') as f:
                return f.read()

    @property
    def text(self):
        """
        Contents of the object as a Unicode string.

        Returns None if this S3Object has not been downloaded.
        """
        if self._path:
            return self.blob.decode('utf-8', errors='replace')

    @property
    def size(self):
        """
        Size of the object in bytes.

        Returns None if the key does not correspond to an object in S3.
        """
        return self._size

    @property
    def has_info(self):
        """
        Returns true if this S3Object contains the content-type or user-metadata.
        If False, this means that content_type and range_info will not return the
        proper information
        """
        return self._content_type is not None or self._metadata is not None

    @property
    def metadata(self):
        """
        Returns a dictionary of user-defined metadata
        """
        return self._metadata

    @property
    def content_type(self):
        """
        Returns the content-type of the S3 object; if unknown, returns None
        """
        return self._content_type

    @property
    def range_info(self):
        """
        Returns a namedtuple containing the following fields:
            - total_size: size in S3 of the object
            - request_offset: the starting offset in this S3Object
            - request_length: the length in this S3Object
        """
        return self._range_info

    def __str__(self):
        if self._path:
            return '<S3Object %s (%d bytes, local)>' % (self._url, self._size)
        elif self._size:
            return '<S3Object %s (%d bytes, in S3)>' % (self._url, self._size)
        else:
            return '<S3Object %s (object does not exist)>' % self._url

    def __repr__(self):
        return str(self)

class S3(object):

    @classmethod
    def get_root_from_config(cls, echo, create_on_absent=True):
        return DATATOOLS_S3ROOT

    def __init__(self,
                 tmproot='.',
                 bucket=None,
                 prefix=None,
                 run=None,
                 s3root=None):
        """
        Initialize a new context for S3 operations. This object is used as
        a context manager for a with statement.

        There are two ways to initialize this object depending whether you want
        to bind paths to a Metaflow run or not.

        1. With a run object:

            run: (required) Either a FlowSpec object (typically 'self') or a
                 Run object corresponding to an existing Metaflow run. These
                 are used to add a version suffix in the S3 path.
            bucket: (optional) S3 bucket.
            prefix: (optional) S3 prefix.

        2. Without a run object:

            s3root: (optional) An S3 root URL for all operations. If this is
                    not specified, all operations require a full S3 URL.

        These options are supported in both the modes:

            tmproot: (optional) Root path for temporary files (default: '.')
        """

        if not boto_found:
            raise MetaflowException("You need to install 'boto3' in order to use S3.")

        if run:
            # 1. use a (current) run ID with optional customizations
            parsed = urlparse(DATATOOLS_S3ROOT)
            if not bucket:
                bucket = parsed.netloc
            if not prefix:
                prefix = parsed.path
            if isinstance(run, FlowSpec):
                if current.is_running_flow:
                    prefix = os.path.join(prefix,
                                          current.flow_name,
                                          current.run_id)
                else:
                    raise MetaflowS3URLException(\
                        "Initializing S3 with a FlowSpec outside of a running "
                        "flow is not supported.")
            else:
                prefix = os.path.join(prefix, run.parent.id, run.id)

            self._s3root = u's3://%s' % os.path.join(bucket, prefix.strip('/'))
        elif s3root:
            # 2. use an explicit S3 prefix
            parsed = urlparse(to_unicode(s3root))
            if parsed.scheme != 's3':
                raise MetaflowS3URLException(\
                    "s3root needs to be an S3 URL prefxied with s3://.")
            self._s3root = s3root.rstrip('/')
        else:
            # 3. use the client only with full URLs
            self._s3root = None

        self._s3_client = None
        self._tmpdir = mkdtemp(dir=tmproot, prefix='metaflow.s3.')

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        """
        Delete all temporary files downloaded in this context.
        """
        try:
            if not debug.s3client:
                if self._tmpdir:
                    shutil.rmtree(self._tmpdir)
                    self._tmpdir = None
        except:
            pass

    def reset_client(self, hard_reset=False):
        if hard_reset or self._s3_client is None:
            from metaflow.datastore.util.s3util import get_s3_client
            self._s3_client, self._s3_client_error = get_s3_client()

    def _url(self, key_value):
        # NOTE: All URLs are handled as Unicode objects (unicode in py2,
        # string in py3) internally. We expect that all URLs passed to this
        # class as either Unicode or UTF-8 encoded byte strings. All URLs
        # returned are Unicode.
        key = getattr(key_value, 'key', key_value)
        if self._s3root is None:
            parsed = urlparse(to_unicode(key))
            if parsed.scheme == 's3' and parsed.path:
                return key
            else:
                if current.is_running_flow:
                    raise MetaflowS3URLException(\
                        "Specify S3(run=self) when you use S3 inside a running "
                        "flow. Otherwise you have to use S3 with full "
                        "s3:// urls.")
                else:
                    raise MetaflowS3URLException(\
                        "Initialize S3 with an 's3root' or 'run' if you don't "
                        "want to specify full s3:// urls.")
        elif key:
            if key.startswith('s3://'):
                raise MetaflowS3URLException(\
                    "Don't use absolute S3 URLs when the S3 client is "
                    "initialized with a prefix. URL: %s" % key)
            return os.path.join(self._s3root, key)
        else:
            return self._s3root

    def _url_and_range(self, key_value):
        url = self._url(key_value)
        start = getattr(key_value, 'offset', None)
        length = getattr(key_value, 'length', None)
        range_str = None
        # Range specification are inclusive so getting from offset 500 for 100
        # bytes will read as bytes=500-599
        if start is not None or length is not None:
            if start is None:
                start = 0
            if length is None:
                # Fetch from offset till the end of the file
                range_str = "bytes=%d-" % start
            elif length < 0:
                # Fetch from end; ignore start value here
                range_str = "bytes=-%d" % (-length)
            else:
                # Typical range fetch
                range_str = "bytes=%d-%d" % (start, start + length - 1)
        return url, range_str


    def list_paths(self, keys=None):
        """
        List the next level of paths in S3. If multiple keys are
        specified, listings are done in parallel. The returned
        S3Objects have .exists == False if the url refers to a
        prefix, not an existing S3 object.

        Args:
            keys: (required) a list of suffixes for paths to list.

        Returns:
            a list of S3Objects (not downloaded)

        Example:

        Consider the following paths in S3:

        A/B/C
        D/E

        In this case, list_paths(['A', 'D']), returns ['A/B', 'D/E']. The
        first S3Object has .exists == False, since it does not refer to an
        object in S3. It is just a prefix.
        """
        def _list(keys):
            if keys is None:
                keys = [None]
            urls = ((self._url(key).rstrip('/') + '/', None) for key in keys)
            res = self._read_many_files('list', urls)
            for s3prefix, s3url, size in res:
                if size:
                    yield s3prefix, s3url, None, int(size)
                else:
                    yield s3prefix, s3url, None, None

        return list(starmap(S3Object, _list(keys)))

    def list_recursive(self, keys=None):
        """
        List objects in S3 recursively. If multiple keys are
        specified, listings are done in parallel. The returned
        S3Objects have always .exists == True, since they refer
        to existing objects in S3.

        Args:
            keys: (required) a list of suffixes for paths to list.

        Returns:
            a list of S3Objects (not downloaded)

        Example:

        Consider the following paths in S3:

        A/B/C
        D/E

        In this case, list_recursive(['A', 'D']), returns ['A/B/C', 'D/E'].
        """
        def _list(keys):
            if keys is None:
                keys = [None]
            res = self._read_many_files('list',
                                        map(self._url_and_range, keys),
                                        recursive=True)
            for s3prefix, s3url, size in res:
                yield s3prefix, s3url, None, int(size)
        return list(starmap(S3Object, _list(keys)))

    def info(self, key=None, return_missing=False):
        """
        Get information about a single object from S3

        Args:
            key: (optional) a suffix identifying the object.
            return_missing: (optional, default False) if set to True, do
                            not raise an exception for a missing key but
                            return it as an S3Object with .exists == False.

        Returns:
            an S3Object containing information about the object. The
            downloaded property will be false and exists will indicate whether
            or not the file exists
        """
        url = self._url(key)
        src = urlparse(url)

        def _info(s3, tmp):
            resp = s3.head_object(Bucket=src.netloc, Key=src.path.lstrip('/"'))
            with open('%s' % tmp, mode='w') as f:
                args = {
                    'content_type': resp['ContentType'],
                    'metadata': resp['Metadata'],
                    'size': resp['ContentLength']}
                json.dump(args, f)

        path_info = None
        try:
            path_info = self._one_boto_op(_info, url)
        except MetaflowS3NotFound:
            if return_missing:
                path_info = None
            else:
                raise
        if path_info:
            with open(path_info, 'r') as f:
                info = json.load(f)
            return S3Object(
                self._s3root, url,
                path=None,
                size=info['size'],
                content_type=info['content_type'],
                metadata=info['metadata'])
        return S3Object(self._s3root, url, None)

    def info_many(self, keys, return_missing=False):
        """
        Get information about many objects from S3 in parallel.

        Args:
            keys: (required) a list of suffixes identifying the objects.
            return_missing: (optional, default False) if set to True, do
                            not raise an exception for a missing key but
                            return it as an S3Object with .exists == False.

        Returns:
            a list of S3Objects corresponding to the objects requested. The
            downloaded property will be false and exists will indicate whether
            or not the file exists.
        """
        def _head():
            from . import s3op
            res = self._read_many_files('info',
                                        map(self._url_and_range, keys),
                                        verbose=False,
                                        listing=True)

            for s3prefix, s3url, fname in res:
                if fname:
                    # We have a metadata file to read from
                    with open(os.path.join(self._tmpdir, fname), 'r') as f:
                        info = json.load(f)
                    if info['error'] is not None:
                        # We have an error, we check if it is a missing file
                        if info['error'] == s3op.ERROR_URL_NOT_FOUND:
                            if return_missing:
                                yield self._s3root, s3url, None
                            else:
                                raise MetaflowS3NotFound()
                        elif info['error'] == s3op.ERROR_URL_ACCESS_DENIED:
                            raise MetaflowS3AccessDenied()
                        else:
                            raise MetaflowS3Exception("Got error: %d" % info['error'])
                    else:
                        yield self._s3root, s3url, None, \
                            info['size'], info['content_type'], info['metadata']
                else:
                    # This should not happen; we should always get a response
                    # even if it contains an error inside it
                    raise MetaflowS3Exception("Did not get a response to HEAD")
        return list(starmap(S3Object, _head()))

    def get(self, key=None, return_missing=False, return_info=True):
        """
        Get a single object from S3.

        Args:
            key: (optional) a suffix identifying the object. Can also be
                 an object containing the properties `key`, `offset` and
                 `length` to specify a range query. `S3GetObject` is such an object.
            return_missing: (optional, default False) if set to True, do
                            not raise an exception for a missing key but
                            return it as an S3Object with .exists == False.
            return_info: (optional, default True) if set to True, fetch the
                         content-type and user metadata associated with the object.

        Returns:
            an S3Object corresponding to the object requested.
        """
        url, r = self._url_and_range(key)
        src = urlparse(url)

        def _download(s3, tmp):
            if r:
                resp = s3.get_object(
                    Bucket=src.netloc,
                    Key=src.path.lstrip('/'),
                    Range=r)
                code = str(resp['ResponseMetadata']['HTTPStatusCode'])
                if code[0] == '2':
                    with open(tmp, mode='w') as t:
                        t.write(resp['Body'].read())
                else:
                    # TODO: Better raised error
                    raise RuntimeError("Could not load file")
            else:
                s3.download_file(src.netloc, src.path.lstrip('/'), tmp)
                return url

        def _info(s3, tmp):
            resp = s3.head_object(Bucket=src.netloc, Key=src.path.lstrip('/"'))
            with open('%s' % tmp, mode='w') as f:
                args = {
                    'content_type': resp['ContentType'],
                    'metadata': resp['Metadata']}
                json.dump(args, f)

        path_info = None
        try:
            path = self._one_boto_op(_download, url)
            if return_info:
                path_info = self._one_boto_op(_info, url)
        except MetaflowS3NotFound:
            if return_missing:
                path = None
            else:
                raise
        if path_info:
            with open(path_info, 'r') as f:
                info = json.load(f)
            return S3Object(
                self._s3root, url, path,
                content_type=info['content_type'],
                metadata=info['metadata'])
        return S3Object(self._s3root, url, path)

    def get_many(self, keys, return_missing=False, return_info=True):
        """
        Get many objects from S3 in parallel.

        Args:
            keys: (required) a list of suffixes identifying the objects. Each
                  item in the list can also be an object containing the properties
                  `key`, `offset` and `length to specify a range query.
                  `S3GetObject` is such an object.
            return_missing: (optional, default False) if set to True, do
                            not raise an exception for a missing key but
                            return it as an S3Object with .exists == False.
            return_info: (optional, default True) if set to True, fetch the
                         content-type and user metadata associated with the object.

        Returns:
            a list of S3Objects corresponding to the objects requested.
        """
        def _get():
            res = self._read_many_files('get',
                                        map(self._url_and_range, keys),
                                        allow_missing=return_missing,
                                        verify=True,
                                        verbose=False,
                                        info=return_info,
                                        listing=True)

            for s3prefix, s3url, fname in res:
                if return_info:
                    if fname:
                        # We have a metadata file to read from
                        with open(os.path.join(self._tmpdir, '%s_meta' % fname),
                                'r') as f:
                            info = json.load(f)
                        yield self._s3root, s3url, os.path.join(self._tmpdir, fname), \
                            None, info['content_type'], info['metadata']
                    else:
                        yield self._s3root, s3prefix, None
                else:
                    if fname:
                        yield self._s3root, s3url, os.path.join(self._tmpdir, fname)
                    else:
                        # missing entries per return_missing=True
                        yield self._s3root, s3prefix, None
        return list(starmap(S3Object, _get()))

    def get_recursive(self, keys, return_info=False):
        """
        Get many objects from S3 recursively in parallel.

        Args:
            keys: (required) a list of suffixes for paths to download
                  recursively.
            return_info: (optional, default False) if set to True, fetch the
                         content-type and user metadata associated with the object.

        Returns:
            a list of S3Objects corresponding to the objects requested.
        """
        def _get():
            res = self._read_many_files('get',
                                        map(self._url_and_range, keys),
                                        recursive=True,
                                        verify=True,
                                        verbose=False,
                                        info=return_info,
                                        listing=True)

            for s3prefix, s3url, fname in res:
                if return_info:
                    # We have a metadata file to read from
                    with open(os.path.join(self._tmpdir, '%s_meta' % fname),
                            'r') as f:
                        info = json.load(f)
                    yield self._s3root, s3url, os.path.join(self._tmpdir, fname), \
                        None, info['content_type'], info['metadata']
                else:
                    yield s3prefix, s3url, os.path.join(self._tmpdir, fname)
        return list(starmap(S3Object, _get()))

    def get_all(self, return_info=False):
        """
        Get all objects from S3 recursively (in parallel). This request
        only works if S3 is initialized with a run or a s3root prefix.

        Args:
            return_info: (optional, default False) if set to True, fetch the
                         content-type and user metadata associated with the object.

        Returns:
            a list of S3Objects corresponding to the objects requested.
        """
        if self._s3root is None:
            raise MetaflowS3URLException(\
                "Can't get_all() when S3 is initialized without a prefix")
        else:
            return self.get_recursive([None], return_info)

    def put(self, key, obj, overwrite=True, content_type=None, metadata=None):
        """
        Put an object to S3.

        Args:
            key:           (required) suffix for the object.
            obj:           (required) a bytes, string, or a unicode object to
                           be stored in S3.
            overwrite:     (optional) overwrites the key with obj, if it exists
            content_type:  (optional) string representing the MIME type of the
                           object
            metadata:      (optional) User metadata to store alongside the object

        Returns:
            an S3 URL corresponding to the object stored.
        """
        if isinstance(obj, (RawIOBase, BufferedIOBase)):
            if not obj.readable() or not obj.seekable():
                raise MetaflowS3InvalidObject(
                    "Object corresponding to the key '%s' is not readable or seekable" %
                    key)
            blob = obj
        else:
            if not is_stringish(obj):
                raise MetaflowS3InvalidObject(\
                    "Object corresponding to the key '%s' is not a string "
                    "or a bytes object." % key)
            blob = to_fileobj(obj)
        # We override the close functionality to prevent closing of the
        # file if it is used multiple times when uploading (since upload_fileobj
        # will/may close it on failure)
        real_close = blob.close
        blob.close = lambda: None

        url = self._url(key)
        src = urlparse(url)
        extra_args = None
        if content_type or metadata:
            extra_args = {}
            if content_type:
                extra_args['ContentType'] = content_type
            if metadata:
                extra_args['Metadata'] = {
                    'metaflow-user-attributes': json.dumps(metadata)}
        
        def _upload(s3, _):
            # We make sure we are at the beginning in case we are retrying
            blob.seek(0)
            s3.upload_fileobj(
                blob, src.netloc, src.path.lstrip('/'), ExtraArgs=extra_args)

        if overwrite:
            self._one_boto_op(_upload, url)
            real_close()
            return url
        else:
            def _head(s3, _):
                s3.head_object(Bucket=src.netloc, Key=src.path.lstrip('/'))

            try:
                self._one_boto_op(_head, url)
            except MetaflowS3NotFound:
                self._one_boto_op(_upload, url)
            finally:
                real_close()
            return url

    def put_many(self, key_objs, overwrite=True):
        """
        Put objects to S3 in parallel.

        Args:
            key_objs:  (required) an iterator of (key, value) tuples. Value must
                       be a string, bytes, or a unicode object. Instead of
                       (key, value) tuples, you can also pass any object that
                       has the following properties 'key', 'value', 'content_type',
                       'metadata' like the S3PutObject for example. 'key' and
                       'value' are required but others are optional.
            overwrite: (optional) overwrites the key with obj, if it exists

        Returns:
            a list of (key, S3 URL) tuples corresponding to the files sent.
        """
        def _store():
            for key_obj in key_objs:
                if isinstance(key_obj, tuple):
                    key = key_obj[0]
                    obj = key_obj[1]
                else:
                    key = key_obj.key
                    obj = key_obj.value
                store_info = {
                    'key': key,
                    'content_type': getattr(key_obj, 'content_type', None)
                }
                metadata = getattr(key_obj, 'metadata', None)
                if metadata:
                    store_info['metadata'] = {
                        'metaflow-user-attributes': json.dumps(metadata)}
                if isinstance(obj, (RawIOBase, BufferedIOBase)):
                    if not obj.readable() or not obj.seekable():
                        raise MetaflowS3InvalidObject(
                            "Object corresponding to the key '%s' is not readable or seekable" %
                            key)
                else:
                    if not is_stringish(obj):
                        raise MetaflowS3InvalidObject(\
                            "Object corresponding to the key '%s' is not a string "
                            "or a bytes object." % key)
                    obj = to_fileobj(obj)
                with NamedTemporaryFile(dir=self._tmpdir,
                                        delete=False,
                                        mode='wb',
                                        prefix='metaflow.s3.put_many.') as tmp:
                    tmp.write(obj.read())
                    tmp.close()
                    yield tmp.name, self._url(key), store_info

        return self._put_many_files(_store(), overwrite)


    def put_files(self, key_paths, overwrite=True):
        """
        Put files to S3 in parallel.

        Args:
            key_paths: (required) an iterator of (key, path) tuples. Instead of
                       (key, path) tuples, you can also pass any object that
                       has the following properties 'key', 'path', 'content_type',
                       'metadata' like the S3PutObject for example. 'key' and
                       'path' are required but others are optional.
            overwrite: (optional) overwrites the key with obj, if it exists

        Returns:
            a list of (key, S3 URL) tuples corresponding to the files sent.
        """
        def _check():
            for key_path in key_paths:
                if isinstance(key_path, tuple):
                    key = key_path[0]
                    path = key_path[1]
                else:
                    key = key_path.key
                    path = key_path.path
                store_info = {
                    'key': key,
                    'content_type': getattr(key_path, 'content_type', None),
                }
                metadata = getattr(key_path, 'metadata', None)
                if metadata:
                    store_info['metadata'] = {
                        'metaflow-user-attributes': json.dumps(metadata)}
                if not os.path.exists(path):
                    raise MetaflowS3NotFound("Local file not found: %s" % path)
                yield path, self._url(key), store_info

        return self._put_many_files(_check(), overwrite)

    def _one_boto_op(self, op, url):
        from . import s3op
        error = ''
        for i in range(NUM_S3OP_RETRIES + 1):
            tmp = NamedTemporaryFile(dir=self._tmpdir,
                                     prefix='metaflow.s3.one_file.',
                                     delete=False)
            self.reset_client()
            try:    
                op(self._s3_client, tmp.name)
                return tmp.name
            except self._s3_client_error as err:
                error_code = s3op.normalize_client_error(err)
                if error_code == 404:
                    raise MetaflowS3NotFound(url)
                elif error_code == 403:
                    raise MetaflowS3AccessDenied(url)
                elif error_code == 'NoSuchBucket':
                    raise MetaflowS3URLException("Specified S3 bucket doesn't exist.")
                error = str(err)
            except Exception as ex:
                # TODO specific error message for out of disk space
                error = str(ex)
            os.unlink(tmp.name)
            self.reset_client(hard_reset = True)
            # add some jitter to make sure retries are not synchronized
            time.sleep(2**i + random.randint(0, 10))
        raise MetaflowS3Exception("S3 operation failed.\n"\
                                  "Key requested: %s\n"\
                                  "Error: %s" % (url, error))

    # NOTE: re: _read_many_files and _put_many_files
    # All file IO is through binary files - we write bytes, we read
    # bytes. All inputs and outputs from these functions are Unicode.
    # Conversion between bytes and unicode is done through 
    # and url_unquote.
    def _read_many_files(self, op, prefixes_and_ranges, **options):
        prefixes_and_ranges = list(prefixes_and_ranges)
        with NamedTemporaryFile(dir=self._tmpdir,
                                mode='wb',
                                delete=not debug.s3client,
                                prefix='metaflow.s3.inputs.') as inputfile:
            inputfile.write(b'\n'.join(
                [b' '.join([url_quote(prefix)] + ([url_quote(r)] if r else []))
                    for prefix, r in prefixes_and_ranges]))
            inputfile.flush()
            stdout, stderr = self._s3op_with_retries(op,
                                                     inputs=inputfile.name,
                                                     **options)
            if stderr:
                raise MetaflowS3Exception("Getting S3 files failed.\n"\
                                          "First prefix requested: %s\n"\
                                          "Error: %s" % (prefixes_and_ranges[0], stderr))
            else:
                for line in stdout.splitlines():
                    yield tuple(map(url_unquote, line.strip(b'\n').split(b' ')))

    def _put_many_files(self, url_info, overwrite):
        url_info = list(url_info)
        url_dicts = [dict(
            chain([
                ('local', os.path.realpath(local)),
                ('url', url)], info.items())) for local, url, info in url_info]

        with NamedTemporaryFile(dir=self._tmpdir,
                                mode='wb',
                                delete=not debug.s3client,
                                prefix='metaflow.s3.put_inputs.') as inputfile:
            lines = [to_bytes(json.dumps(x)) for x in url_dicts]
            inputfile.write(b'\n'.join(lines))
            inputfile.flush()
            stdout, stderr = self._s3op_with_retries('put',
                                                filelist=inputfile.name,
                                                verbose=False,
                                                overwrite=overwrite,
                                                listing=True)
            if stderr:
                raise MetaflowS3Exception("Uploading S3 files failed.\n"\
                                          "First key: %s\n"\
                                          "Error: %s" % (url_info[0][2]['key'],
                                                         stderr))
            else:
                urls = set()
                for line in stdout.splitlines():
                    url, _, _ = map(url_unquote, line.strip(b'\n').split(b' '))
                    urls.add(url)
                return [(info['key'], url) for _, url, info in url_info if url in urls]

    def _s3op_with_retries(self, mode, **options):
        from . import s3op
        cmdline = [sys.executable, os.path.abspath(s3op.__file__), mode]
        for key, value in options.items():
            key = key.replace('_', '-')
            if isinstance(value, bool):
                if value:
                    cmdline.append('--%s' % key)
                else:
                    cmdline.append('--no-%s' % key)
            else:
                cmdline.extend(('--%s' % key, value))

        for i in range(NUM_S3OP_RETRIES + 1):
            with NamedTemporaryFile(dir=self._tmpdir,
                                    mode='wb+',
                                    delete=not debug.s3client,
                                    prefix='metaflow.s3op.stderr') as stderr:
                try:
                    debug.s3client_exec(cmdline)
                    stdout = subprocess.check_output(cmdline,
                                                     cwd=self._tmpdir,
                                                     stderr=stderr.file)
                    return stdout, None
                except subprocess.CalledProcessError as ex:
                    stderr.seek(0)
                    err_out = stderr.read().decode('utf-8', errors='replace')
                    stderr.seek(0)
                    if ex.returncode == s3op.ERROR_URL_NOT_FOUND:
                        raise MetaflowS3NotFound(err_out)
                    elif ex.returncode == s3op.ERROR_URL_ACCESS_DENIED:
                        raise MetaflowS3AccessDenied(err_out)
                    time.sleep(2**i + random.randint(0, 10))

        return None, err_out



import os
from collections import defaultdict
import sys
import hashlib
import json
import time
import string
import random
import uuid

from metaflow.exception import MetaflowException, MetaflowInternalError
from metaflow.plugins import ResourcesDecorator, BatchDecorator, RetryDecorator
from metaflow.parameters import deploy_time_eval
from metaflow.decorators import flow_decorators
from metaflow.util import compress_list, dict_to_cli_options, to_pascalcase
from metaflow.metaflow_config import SFN_IAM_ROLE, \
    EVENTS_SFN_ACCESS_IAM_ROLE, SFN_DYNAMO_DB_TABLE, SFN_EXECUTION_LOG_GROUP_ARN
from metaflow import R

from .step_functions_client import StepFunctionsClient
from .event_bridge_client import EventBridgeClient
from ..batch.batch import Batch


class StepFunctionsException(MetaflowException):
    headline = 'AWS Step Functions error'

class StepFunctionsSchedulingException(MetaflowException):
    headline = 'AWS Step Functions scheduling error'

class StepFunctions(object):

    def __init__(self,
                 name,
                 graph,
                 flow,
                 code_package,
                 code_package_url,
                 production_token,
                 metadata,
                 datastore,
                 environment,
                 event_logger,
                 monitor,
                 tags=None,
                 namespace=None,
                 username=None,
                 max_workers=None,
                 workflow_timeout=None,
                 is_project=False):
        self.name = name
        self.graph = graph
        self.flow = flow
        self.code_package = code_package
        self.code_package_url = code_package_url
        self.production_token = production_token
        self.metadata = metadata
        self.datastore = datastore
        self.environment = environment
        self.event_logger = event_logger
        self.monitor = monitor
        self.tags = tags
        self.namespace = namespace
        self.username = username
        self.max_workers = max_workers
        self.workflow_timeout = workflow_timeout
        
        self._client = StepFunctionsClient()
        self._workflow = self._compile()
        self._cron = self._cron()
        self._state_machine_arn = None

    def to_json(self):
        return self._workflow.to_json(pretty=True)

    def trigger_explanation(self):
        if self._cron:
            # Sometime in the future, we should vendor (or write) a utility
            # that can translate cron specifications into a human readable 
            # format and push to the user for a better UX, someday.
            return 'This workflow triggers automatically '\
                'via a cron schedule *%s* defined in AWS EventBridge.' \
                % self.event_bridge_rule
        else:
            return 'No triggers defined. '\
                'You need to launch this workflow manually.'

    def deploy(self, log_execution_history):
        if SFN_IAM_ROLE is None:
            raise StepFunctionsException("No IAM role found for AWS Step "
                                         "Functions. You can create one "
                                         "following the instructions listed at "
                                         "*https://admin-docs.metaflow.org/meta"
                                         "flow-on-aws/deployment-guide/manual-d"
                                         "eployment#scheduling* and "
                                         "re-configure Metaflow using "
                                         "*metaflow configure aws* on your "
                                         "terminal.")
        if log_execution_history:
            if SFN_EXECUTION_LOG_GROUP_ARN is None:
                raise StepFunctionsException("No AWS CloudWatch Logs log "
                                             "group ARN found for emitting "
                                             "state machine execution logs for "
                                             "your workflow. You can set it in "
                                             "your environment by using the "
                                             "METAFLOW_SFN_EXECUTION_LOG_GROUP_ARN "
                                             "environment variable.")
        try:
            self._state_machine_arn = self._client.push(
                    name = self.name, 
                    definition = self.to_json(), 
                    role_arn = SFN_IAM_ROLE,
                    log_execution_history = log_execution_history
                )
        except Exception as e:
            raise StepFunctionsException(repr(e))

    def schedule(self):
        # Scheduling is currently enabled via AWS Event Bridge.
        if EVENTS_SFN_ACCESS_IAM_ROLE is None:
            raise StepFunctionsSchedulingException("No IAM role found for AWS "
                                                   "Events Bridge. You can "
                                                   "create one following the "
                                                   "instructions listed at "
                                                   "*https://admin-docs.metaflo"
                                                   "w.org/metaflow-on-aws/deplo"
                                                   "yment-guide/manual-deployme"
                                                   "nt#scheduling* and "
                                                   "re-configure Metaflow "
                                                   "using *metaflow configure "
                                                   "aws* on your terminal.")
        try:
            self.event_bridge_rule = EventBridgeClient(self.name) \
                                        .cron(self._cron) \
                                        .role_arn(EVENTS_SFN_ACCESS_IAM_ROLE) \
                                        .state_machine_arn(self._state_machine_arn) \
                                        .schedule()
        except Exception as e:
            raise StepFunctionsSchedulingException(repr(e))

    @classmethod
    def trigger(cls, name, parameters):
        try:
            state_machine = StepFunctionsClient().get(name)
        except Exception as e:
            raise StepFunctionsException(repr(e))
        if state_machine is None:
            raise StepFunctionsException("The workflow *%s* doesn't exist "
                                         "on AWS Step Functions. Please "
                                         "deploy your flow first." % name)
        # Dump parameters into `Parameters` input field.
        input = json.dumps({"Parameters" : json.dumps(parameters)})
        # AWS Step Functions limits input to be 32KiB, but AWS Batch
        # has it's own limitation of 30KiB for job specification length.
        # Reserving 10KiB for rest of the job sprecification leaves 20KiB
        # for us, which should be enough for most use cases for now.
        if len(input) > 20480:
            raise StepFunctionsException("Length of parameter names and "
                                         "values shouldn't exceed 20480 as "
                                         "imposed by AWS Step Functions.")
        try:
            state_machine_arn = state_machine.get('stateMachineArn')
            return StepFunctionsClient().trigger(state_machine_arn, input)
        except Exception as e:
            raise StepFunctionsException(repr(e))

    @classmethod
    def list(cls, name, states):
        try:
            state_machine = StepFunctionsClient().get(name)
        except Exception as e:
            raise StepFunctionsException(repr(e))
        if state_machine is None:
            raise StepFunctionsException("The workflow *%s* doesn't exist "
                                         "on AWS Step Functions." % name)
        try:
            state_machine_arn = state_machine.get('stateMachineArn')
            return StepFunctionsClient() \
                .list_executions(state_machine_arn, states)
        except Exception as e:
            raise StepFunctionsException(repr(e))

    @classmethod
    def get_existing_deployment(cls, name):
        workflow = StepFunctionsClient().get(name)
        if workflow is not None:
            try:
                start = json.loads(workflow['definition'])['States']['start']
                parameters = start['Parameters']['Parameters']
                return parameters.get('metaflow.owner'), \
                    parameters.get('metaflow.production_token')
            except KeyError as e:
                raise StepFunctionsException("An exisiting non-metaflow "
                                             "workflow with the same name as "
                                             "*%s* already exists in AWS Step "
                                             "Functions. Please modify the "
                                             "name of this flow or delete your "
                                             "existing workflow on AWS Step "
                                             "Functions." % name)
        return None

    def _compile(self):

        # Visit every node of the flow and recursively build the state machine.
        def _visit(node, workflow, exit_node=None):
            # Assign an AWS Batch job to the AWS Step Functions state
            # and pass the intermediate state by exposing `JobId` and 
            # `Parameters` to the child job(s) as outputs. `Index` and 
            # `SplitParentTaskId` are populated optionally, when available.

            # We can't modify the names of keys in AWS Step Functions aside
            # from a blessed few which are set as `Parameters` for the Map
            # state. That's why even though `JobId` refers to the parent task 
            # id, we can't call it as such. Similar situation for `Parameters`.
            state = State(node.name) \
                        .batch(self._batch(node)) \
                        .output_path('$.[\'JobId\', '
                                    '\'Parameters\', '
                                    '\'Index\', '
                                    '\'SplitParentTaskId\']')
            # End the (sub)workflow if we have reached the end of the flow or
            # the parent step of matching_join of the sub workflow.
            if node.type == 'end' or exit_node in node.out_funcs:
                workflow.add_state(state.end())
            # Continue linear assignment within the (sub)workflow if the node 
            # doesn't branch or fork.
            elif node.type in ('linear', 'join'):
                workflow.add_state(state.next(node.out_funcs[0]))
                _visit(self.graph[node.out_funcs[0]], workflow, exit_node)
            # Create a `Parallel` state and assign sub workflows if the node
            # branches out.
            elif node.type == 'split-and':
                branch_name = hashlib.sha224('&'.join(node.out_funcs) \
                                     .encode('utf-8')) \
                                     .hexdigest()
                workflow.add_state(state.next(branch_name))
                branch = Parallel(branch_name) \
                            .next(node.matching_join)
                # Generate as many sub workflows as branches and recurse.
                for n in node.out_funcs:
                    branch.branch(
                        _visit(
                            self.graph[n], 
                            Workflow(n).start_at(n), 
                            node.matching_join))
                workflow.add_state(branch)
                # Continue the traversal from the matching_join.
                _visit(self.graph[node.matching_join], workflow, exit_node)
            # Create a `Map` state and assign sub workflow if the node forks.
            elif node.type == 'foreach':
                # Fetch runtime cardinality via an AWS DynamoDb Get call before
                # configuring the node
                cardinality_state_name = '#%s' % node.out_funcs[0]
                workflow.add_state(state.next(cardinality_state_name))
                cardinality_state = State(cardinality_state_name) \
                                        .dynamo_db(SFN_DYNAMO_DB_TABLE, 
                                            '$.JobId', 
                                            'for_each_cardinality') \
                                        .result_path('$.Result')
                iterator_name = '*%s' % node.out_funcs[0]
                workflow.add_state(cardinality_state.next(iterator_name))
                workflow.add_state(
                    Map(iterator_name) \
                        .items_path('$.Result.Item.for_each_cardinality.NS') \
                        .parameter('JobId.$', '$.JobId') \
                        .parameter('SplitParentTaskId.$', '$.JobId') \
                        .parameter('Parameters.$', '$.Parameters') \
                        .parameter('Index.$', '$$.Map.Item.Value') \
                        .next(node.matching_join) \
                        .iterator(
                            _visit(
                                self.graph[node.out_funcs[0]], 
                                Workflow(node.out_funcs[0]) \
                                    .start_at(node.out_funcs[0]), 
                                node.matching_join)) \
                        .max_concurrency(self.max_workers) \
                        .output_path('$.[0]'))
                # Continue the traversal from the matching_join.
                _visit(self.graph[node.matching_join], workflow, exit_node)
            # We shouldn't ideally ever get here.
            else:
                raise StepFunctionsException("Node type *%s* for  step *%s* "
                                             "is not currently supported by "
                                             "AWS Step Functions." 
                                             % (node.type, node.name))
            return workflow

        workflow = Workflow(self.name) \
                        .start_at('start')
        if self.workflow_timeout:
            workflow.timeout_seconds(self.workflow_timeout)
        return _visit(self.graph['start'], workflow)

    def _cron(self):
        schedule = self.flow._flow_decorators.get('schedule')
        if schedule:
            return schedule.schedule
        return None

    def _process_parameters(self):
        parameters = []
        has_schedule = self._cron() is not None
        seen = set()
        for var, param in self.flow._get_parameters():
            # Throw an exception if the parameter is specified twice.
            norm = param.name.lower()
            if norm in seen:
                raise MetaflowException("Parameter *%s* is specified twice. "
                                        "Note that parameter names are "
                                        "case-insensitive." % param.name)
            seen.add(norm)

            is_required = param.kwargs.get('required', False)
            # Throw an exception if a schedule is set for a flow with required
            # parameters with no defaults. We currently don't have any notion
            # of data triggers in AWS Event Bridge.
            if 'default' not in param.kwargs and is_required and has_schedule:
                raise MetaflowException("The parameter *%s* does not have a "
                                        "default and is required. Scheduling "
                                        "such parameters via AWS Event Bridge "
                                        "is not currently supported."
                                        % param.name)
            value = deploy_time_eval(param.kwargs.get('default'))
            parameters.append(dict(name=param.name,
                                   value=value))
        return parameters

    def _batch(self, node):
        attrs = {
            # metaflow.user is only used for setting the AWS Job Name.
            # Since job executions are no longer tied to a specific user
            # identity, we will just set their user to `SFN`. We still do need
            # access to the owner of the workflow for production tokens, which 
            # we can stash in metaflow.owner.
            'metaflow.user': 'SFN',
            'metaflow.owner': self.username,
            'metaflow.flow_name': self.flow.name,
            'metaflow.step_name': node.name,
            'metaflow.run_id.$': '$$.Execution.Name',
            # Unfortunately we can't set the task id here since AWS Step 
            # Functions lacks any notion of run-scoped task identifiers. We 
            # instead co-opt the AWS Batch job id as the task id. This also
            # means that the AWS Batch job name will have missing fields since
            # the job id is determined at job execution, but since the job id is
            # part of the job description payload, we don't lose much except for
            # a few ugly looking black fields in the AWS Batch UI.
            
            # Also, unfortunately we can't set the retry count since 
            # `$$.State.RetryCount` resolves to an int dynamically and 
            # AWS Batch job specification only accepts strings. We handle 
            # retries/catch within AWS Batch to get around this limitation.
            'metaflow.version': self.environment.get_environment_info()[
                'metaflow_version'
            ],
            # We rely on step names and task ids of parent steps to construct
            # input paths for a task. Since the only information we can pass 
            # between states (via `InputPath` and `ResultPath`) in AWS Step 
            # Functions is the job description, we run the risk of exceeding 
            # 32K state size limit rather quickly if we don't filter the job 
            # description to a minimal set of fields. Unfortunately, the partial
            # `JsonPath` implementation within AWS Step Functions makes this 
            # work a little non-trivial; it doesn't like dots in keys, so we
            # have to add the field again. 
            # This pattern is repeated in a lot of other places, where we use 
            # AWS Batch parameters to store AWS Step Functions state 
            # information, since this field is the only field in the AWS Batch 
            # specification that allows us to set key-values.
            'step_name': node.name
        }

        # Store production token within the `start` step, so that subsequent
        # `step-functions create` calls can perform a rudimentary authorization
        # check.
        if node.name == 'start':
            attrs['metaflow.production_token'] = self.production_token

        # Add env vars from the optional @environment decorator.
        env_deco = [deco for deco in node.decorators 
                        if deco.name == 'environment']
        env = {}
        if env_deco:
            env = env_deco[0].attributes['vars']

        if node.name == 'start':
            # Initialize parameters for the flow in the `start` step.
            parameters = self._process_parameters()
            if parameters:
                # Get user-defined parameters from State Machine Input.
                # Since AWS Step Functions doesn't allow for optional inputs
                # currently, we have to unfortunately place an artificial 
                # constraint that every parameterized workflow needs to include
                # `Parameters` as a key in the input to the workflow. 
                # `step-functions trigger` already takes care of this 
                # requirement, but within the UI, the users will be required to
                # specify an input with key as `Parameters` and value as a
                # stringified json of the actual parameters - 
                # {"Parameters": "{\"alpha\": \"beta\"}"}
                env['METAFLOW_PARAMETERS'] = '$.Parameters'
                default_parameters = {}
                for parameter in parameters:
                    if parameter['value'] is not None:
                        default_parameters[parameter['name']] = \
                            parameter['value']
                # Dump the default values specified in the flow.
                env['METAFLOW_DEFAULT_PARAMETERS'] = \
                    json.dumps(default_parameters)
            # `start` step has no upstream input dependencies aside from
            # parameters.
            input_paths = None
        else:
            # We need to rely on the `InputPath` of the AWS Step Functions
            # specification to grab task ids and the step names of the parent
            # to properly construct input_paths at runtime. Thanks to the 
            # JsonPath-foo embedded in the parent states, we have this 
            # information easily available. 

            # Handle foreach join.
            if node.type == 'join' and \
                 self.graph[node.split_parents[-1]].type == 'foreach':
                input_paths = \
                    'sfn-${METAFLOW_RUN_ID}/%s/:' \
                        '${METAFLOW_PARENT_TASK_IDS}' % node.in_funcs[0]
                # Unfortunately, AWS Batch only allows strings as value types
                # in it's specification and we don't have any way to concatenate
                # the task ids array from the parent steps within AWS Step 
                # Functions and pass it down to AWS Batch. We instead have to 
                # rely on publishing the state to DynamoDb and fetching it back
                # in within the AWS Batch entry point to set
                # `METAFLOW_PARENT_TASK_IDS`. The state is scoped to the parent
                # foreach task `METAFLOW_SPLIT_PARENT_TASK_ID`. We decided on
                # AWS DynamoDb and not AWS Lambdas, because deploying and 
                # debugging Lambdas would be a nightmare as far as OSS support 
                # is concerned.
                env['METAFLOW_SPLIT_PARENT_TASK_ID'] = \
                    '$.Parameters.split_parent_task_id_%s' % \
                        node.split_parents[-1]
            else:
                # Set appropriate environment variables for runtime replacement.
                if len(node.in_funcs) == 1:
                    input_paths = \
                        'sfn-${METAFLOW_RUN_ID}/%s/${METAFLOW_PARENT_TASK_ID}' \
                            % node.in_funcs[0]
                    env['METAFLOW_PARENT_TASK_ID'] = '$.JobId'
                else:
                    # Generate the input paths in a quasi-compressed format.
                    # See util.decompress_list for why this is written the way 
                    # it is.
                    input_paths = 'sfn-${METAFLOW_RUN_ID}:' + ','.join(
                        '/${METAFLOW_PARENT_%s_STEP}/'
                            '${METAFLOW_PARENT_%s_TASK_ID}' % (idx, idx) 
                            for idx, _ in enumerate(node.in_funcs))
                    for idx, _ in enumerate(node.in_funcs):
                        env['METAFLOW_PARENT_%s_TASK_ID' % idx] = \
                            '$.[%s].JobId' % idx
                        env['METAFLOW_PARENT_%s_STEP' % idx] = \
                            '$.[%s].Parameters.step_name' % idx
            env['METAFLOW_INPUT_PATHS'] = input_paths

            if node.is_inside_foreach:
                # Set the task id of the parent job of the foreach split in 
                # our favorite dumping ground, the AWS Batch attrs. For 
                # subsequent descendent tasks, this attrs blob becomes the 
                # input to those descendent tasks. We set and propagate the
                # task ids pointing to split_parents through every state.
                if any(self.graph[n].type == 'foreach' for n in node.in_funcs):
                    attrs['split_parent_task_id_%s.$' % \
                        node.split_parents[-1]] = '$.SplitParentTaskId'
                    for parent in node.split_parents[:-1]:
                        if self.graph[parent].type == 'foreach':
                            attrs['split_parent_task_id_%s.$' % parent] = \
                                '$.Parameters.split_parent_task_id_%s' % parent
                elif node.type == 'join':
                    if self.graph[node.split_parents[-1]].type == 'foreach':
                        # A foreach join only gets one set of input from the
                        # parent tasks. We filter the Map state to only output
                        # `$.[0]`, since we don't need any of the other outputs,
                        # that information is available to us from AWS DynamoDB.
                        # This has a nice side-effect of making our foreach
                        # splits infinitely scalable because otherwise we would
                        # be bounded by the 32K state limit for the outputs. So,
                        # instead of referencing `Parameters` fields by index
                        # (like in `split-and`), we can just reference them
                        # directly.
                        attrs['split_parent_task_id_%s.$' % \
                            node.split_parents[-1]] = \
                                '$.Parameters.split_parent_task_id_%s' % \
                                    node.split_parents[-1]
                        for parent in node.split_parents[:-1]:
                            if self.graph[parent].type == 'foreach':
                                attrs['split_parent_task_id_%s.$' % parent] = \
                                    '$.Parameters.split_parent_task_id_%s' % \
                                        parent
                    else:
                        for parent in node.split_parents:
                            if self.graph[parent].type == 'foreach':
                                attrs['split_parent_task_id_%s.$' % parent] = \
                                '$.[0].Parameters.split_parent_task_id_%s' % \
                                    parent
                else:
                    for parent in node.split_parents:
                        if self.graph[parent].type == 'foreach':
                            attrs['split_parent_task_id_%s.$' % parent] = \
                                '$.Parameters.split_parent_task_id_%s' % parent

                # Set `METAFLOW_SPLIT_PARENT_TASK_ID_FOR_FOREACH_JOIN` if the 
                # next transition is to a foreach join, so that the 
                # stepfunctions decorator can write the mapping for input path
                # to DynamoDb.
                if any(self.graph[n].type == 'join' and \
                        self.graph[self.graph[n].split_parents[-1]].type == \
                        'foreach'
                            for n in node.out_funcs):
                    env['METAFLOW_SPLIT_PARENT_TASK_ID_FOR_FOREACH_JOIN'] = \
                            attrs['split_parent_task_id_%s.$' % \
                                self.graph[node.out_funcs[0]].split_parents[-1]]

                # Set ttl for the values we set in AWS DynamoDB.
                if node.type == 'foreach':
                    if self.workflow_timeout:
                        env['METAFLOW_SFN_WORKFLOW_TIMEOUT'] = \
                            self.workflow_timeout

            # Handle split index for for-each.
            if any(self.graph[n].type == 'foreach' for n in node.in_funcs):
                env['METAFLOW_SPLIT_INDEX'] = '$.Index'

        env['METAFLOW_CODE_URL'] = self.code_package_url
        env['METAFLOW_FLOW_NAME'] = attrs['metaflow.flow_name']
        env['METAFLOW_STEP_NAME'] = attrs['metaflow.step_name']
        env['METAFLOW_RUN_ID'] = attrs['metaflow.run_id.$']
        env['METAFLOW_PRODUCTION_TOKEN'] = self.production_token
        env['SFN_STATE_MACHINE'] = self.name
        env['METAFLOW_OWNER'] = attrs['metaflow.owner']
        # Can't set `METAFLOW_TASK_ID` due to lack of run-scoped identifiers.
        # We will instead rely on `AWS_BATCH_JOB_ID` as the task identifier.
        # Can't set `METAFLOW_RETRY_COUNT` either due to integer casting issue.
        metadata_env = self.metadata.get_runtime_environment('step-functions')
        env.update(metadata_env)

        metaflow_version = self.environment.get_environment_info()
        metaflow_version['flow_name'] = self.graph.name
        metaflow_version['production_token'] = self.production_token
        env['METAFLOW_VERSION'] = json.dumps(metaflow_version)

        # Set AWS DynamoDb Table Name for state tracking for for-eaches.
        # There are three instances when metaflow runtime directly interacts
        # with AWS DynamoDB.
        #   1. To set the cardinality of foreaches (which are subsequently)
        #      read prior to the instantiation of the Map state by AWS Step
        #      Functions.
        #   2. To set the input paths from the parent steps of a foreach join.
        #   3. To read the input paths in a foreach join.
        if node.type == 'foreach' or \
            (node.is_inside_foreach and \
                any(self.graph[n].type == 'join' and \
                    self.graph[self.graph[n].split_parents[-1]].type == \
                        'foreach' for n in node.out_funcs)) or \
            (node.type == 'join' and \
                self.graph[node.split_parents[-1]].type == 'foreach'):
            if SFN_DYNAMO_DB_TABLE is None:
                raise StepFunctionsException("An AWS DynamoDB table is needed "
                                             "to support foreach in your flow. "
                                             "You can create one following the "
                                             "instructions listed at *https://a"
                                             "dmin-docs.metaflow.org/metaflow-o"
                                             "n-aws/deployment-guide/manual-dep"
                                             "loyment#scheduling* and "
                                             "re-configure Metaflow using "
                                             "*metaflow configure aws* on your "
                                             "terminal.")
            env['METAFLOW_SFN_DYNAMO_DB_TABLE'] = SFN_DYNAMO_DB_TABLE

        # Resolve AWS Batch resource requirements.
        batch_deco = [deco for deco in node.decorators
                        if deco.name == 'batch'][0]
        resources = batch_deco.attributes

        # Resolve retry strategy.
        user_code_retries, total_retries= self._get_retries(node)

        task_spec = {
            'flow_name': attrs['metaflow.flow_name'],
            'step_name': attrs['metaflow.step_name'],
            'run_id': 'sfn-$METAFLOW_RUN_ID',
            # Use AWS Batch job identifier as the globally unique 
            # task identifier.
            'task_id': '$AWS_BATCH_JOB_ID',
            # Since retries are handled by AWS Batch, we can rely on
            # AWS_BATCH_JOB_ATTEMPT as the job counter.
            'retry_count': '$((AWS_BATCH_JOB_ATTEMPT-1))'
        }

        return Batch(self.metadata, self.environment) \
                .create_job(
                        step_name=node.name,
                        step_cli=self._step_cli(node,
                                                input_paths,
                                                self.code_package_url,
                                                user_code_retries),
                        task_spec=task_spec,
                        code_package_sha=self.code_package.sha,
                        code_package_url=self.code_package_url,
                        code_package_ds=self.datastore.TYPE,
                        image=resources['image'],
                        queue=resources['queue'],
                        iam_role=resources['iam_role'],
                        execution_role=resources['execution_role'],
                        cpu=resources['cpu'],
                        gpu=resources['gpu'],
                        memory=resources['memory'],
                        run_time_limit=batch_deco.run_time_limit,
                        shared_memory=resources['shared_memory'],
                        max_swap=resources['max_swap'],
                        swappiness=resources['swappiness'],
                        env=env,
                        attrs=attrs
                ) \
                .attempts(total_retries + 1)

    def _get_retries(self, node):
        max_user_code_retries = 0
        max_error_retries = 0
        # Different decorators may have different retrying strategies, so take
        # the max of them.
        for deco in node.decorators:
            user_code_retries, error_retries = deco.step_task_retry_count()
            max_user_code_retries = max(max_user_code_retries,
                                        user_code_retries)
            max_error_retries = max(max_error_retries, error_retries)

        return max_user_code_retries,\
            max_user_code_retries + max_error_retries

    def _step_cli(self, 
                  node,
                  paths, 
                  code_package_url, 
                  user_code_retries):
        cmds = []

        script_name = os.path.basename(sys.argv[0])
        executable = self.environment.executable(node.name)

        if R.use_r():
            entrypoint = [R.entrypoint()]
        else:
            entrypoint = [executable, script_name]

        # Use AWS Batch job identifier as the globally unique task identifier.
        task_id = '${AWS_BATCH_JOB_ID}'

        # FlowDecorators can define their own top-level options. They are
        # responsible for adding their own top-level options and values through
        # the get_top_level_options() hook. See similar logic in runtime.py.
        top_opts_dict = {}
        for deco in flow_decorators():
            top_opts_dict.update(deco.get_top_level_options())
        top_opts = list(dict_to_cli_options(top_opts_dict))

        if node.name == 'start':
            # We need a separate unique ID for the special _parameters task
            task_id_params = '%s-params' % task_id
            # Export user-defined parameters into runtime environment
            param_file = ''.join(random.choice(string.ascii_lowercase) 
                for _ in range(10))
            export_params = \
                'python -m ' \
                'metaflow.plugins.aws.step_functions.set_batch_environment ' \
                'parameters %s && . `pwd`/%s' % (param_file, param_file)
            params = entrypoint + top_opts +\
                ['--quiet',
                 '--metadata=%s' % self.metadata.TYPE,
                 '--environment=%s' % self.environment.TYPE,
                 '--datastore=s3',
                 '--event-logger=%s' % self.event_logger.logger_type,
                 '--monitor=%s' % self.monitor.monitor_type,
                 '--no-pylint',
                 'init',
                 '--run-id sfn-$METAFLOW_RUN_ID',
                 '--task-id %s' % task_id_params]
            # Assign tags to run objects.
            if self.tags:
                params.extend('--tag %s' % tag for tag in self.tags)

            # If the start step gets retried, we must be careful not to 
            # regenerate multiple parameters tasks. Hence we check first if 
            # _parameters exists already.
            exists = entrypoint +\
                ['dump',
                 '--max-value-size=0',
                 'sfn-${METAFLOW_RUN_ID}/_parameters/%s' % (task_id_params)]
            cmd = 'if ! %s >/dev/null 2>/dev/null; then %s && %s; fi'\
                  % (' '.join(exists), export_params, ' '.join(params))
            cmds.append(cmd)
            paths = 'sfn-${METAFLOW_RUN_ID}/_parameters/%s' % (task_id_params)
        
        if node.type == 'join' and\
            self.graph[node.split_parents[-1]].type == 'foreach':
            parent_tasks_file = ''.join(random.choice(string.ascii_lowercase) 
                for _ in range(10))
            export_parent_tasks = \
                'python -m ' \
                'metaflow.plugins.aws.step_functions.set_batch_environment ' \
                'parent_tasks %s && . `pwd`/%s' \
                    % (parent_tasks_file, parent_tasks_file)
            cmds.append(export_parent_tasks)

        top_level = top_opts + [
            '--quiet',
            '--metadata=%s' % self.metadata.TYPE,
            '--environment=%s' % self.environment.TYPE,
            '--datastore=%s' % self.datastore.TYPE,
            '--datastore-root=%s' % self.datastore.datastore_root,
            '--event-logger=%s' % self.event_logger.logger_type,
            '--monitor=%s' % self.monitor.monitor_type,
            '--no-pylint',
            '--with=step_functions_internal'
        ]

        step = [
            'step',
            node.name,
            '--run-id sfn-$METAFLOW_RUN_ID',
            '--task-id %s' % task_id,
            # Since retries are handled by AWS Batch, we can rely on
            # AWS_BATCH_JOB_ATTEMPT as the job counter.
            '--retry-count $((AWS_BATCH_JOB_ATTEMPT-1))',
            '--max-user-code-retries %d' % user_code_retries,
            '--input-paths %s' % paths,
            # Set decorator to batch to execute `task_*` hooks for batch
            # decorator.
            '--with=batch'
        ]
        if any(self.graph[n].type == 'foreach' for n in node.in_funcs):
            # We set the `METAFLOW_SPLIT_INDEX` through JSONPath-foo
            # to pass the state from the parent DynamoDb state for for-each.
            step.append('--split-index $METAFLOW_SPLIT_INDEX')
        if self.tags:
            step.extend('--tag %s' % tag for tag in self.tags)
        if self.namespace is not None:
            step.append('--namespace=%s' % self.namespace)
        cmds.append(' '.join(entrypoint + top_level + step))
        return ' && '.join(cmds)


class Workflow(object):
    def __init__(self, name):
        self.name = name
        tree = lambda: defaultdict(tree)
        self.payload = tree()

    def start_at(self, start_at):
        self.payload['StartAt'] = start_at
        return self

    def add_state(self, state):
        self.payload['States'][state.name] = state.payload
        return self

    def timeout_seconds(self, timeout_seconds):
        self.payload['TimeoutSeconds'] = timeout_seconds
        return self

    def to_json(self, pretty=False):
        return json.dumps(self.payload, indent=4 if pretty else None)


class State(object):

    def __init__(self, name):
        self.name = name
        tree = lambda: defaultdict(tree)
        self.payload = tree()
        self.payload['Type'] = 'Task'

    def resource(self, resource):
        self.payload['Resource'] = resource
        return self

    def next(self, state):
        self.payload['Next'] = state
        return self
        
    def end(self):
        self.payload['End'] = True
        return self

    def parameter(self, name, value):
        self.payload['Parameters'][name] = value
        return self

    def output_path(self, output_path):
        self.payload['OutputPath'] = output_path
        return self

    def result_path(self, result_path):
        self.payload['ResultPath'] = result_path
        return self

    def _partition(self):
        # This is needed to support AWS Gov Cloud and AWS CN regions
        return SFN_IAM_ROLE.split(':')[1]

    def batch(self, job):
        self.resource('arn:%s:states:::batch:submitJob.sync' 
                % self._partition()) \
            .parameter('JobDefinition', job.payload['jobDefinition']) \
            .parameter('JobName', job.payload['jobName']) \
            .parameter('JobQueue', job.payload['jobQueue']) \
            .parameter('Parameters', job.payload['parameters']) \
            .parameter('ContainerOverrides', 
                to_pascalcase(job.payload['containerOverrides'])) \
            .parameter('RetryStrategy', 
                to_pascalcase(job.payload['retryStrategy'])) \
            .parameter('Timeout', 
                to_pascalcase(job.payload['timeout']))
        # tags may not be present in all scenarios
        if 'tags' in job.payload:
            self.parameter('Tags', job.payload['tags'])
        return self

    def dynamo_db(self, table_name, primary_key, values):
        self.resource('arn:%s:states:::dynamodb:getItem' % self._partition()) \
            .parameter('TableName', table_name) \
            .parameter('Key', {
                "pathspec": {
                    "S.$": primary_key
                }
            }) \
            .parameter('ConsistentRead', True) \
            .parameter('ProjectionExpression', values)
        return self  


class Parallel(object):

    def __init__(self, name):
        self.name = name
        tree = lambda: defaultdict(tree)
        self.payload = tree()
        self.payload['Type'] = 'Parallel'

    def branch(self, workflow):
        if 'Branches' not in self.payload:
            self.payload['Branches'] = []
        self.payload['Branches'].append(workflow.payload)
        return self

    def next(self, state):
        self.payload['Next'] = state
        return self

    def output_path(self, output_path):
        self.payload['OutputPath'] = output_path
        return self

    def result_path(self, result_path):
        self.payload['ResultPath'] = result_path
        return self   


class Map(object):

    def __init__(self, name):
        self.name = name
        tree = lambda: defaultdict(tree)
        self.payload = tree()
        self.payload['Type'] = 'Map'
        self.payload['MaxConcurrency'] = 0

    def iterator(self, workflow):
        self.payload['Iterator'] = workflow.payload
        return self

    def next(self, state):
        self.payload['Next'] = state
        return self

    def items_path(self, items_path):
        self.payload['ItemsPath'] = items_path
        return self

    def parameter(self, name, value):
        self.payload['Parameters'][name] = value
        return self

    def max_concurrency(self, max_concurrency):
        self.payload['MaxConcurrency'] = max_concurrency
        return self

    def output_path(self, output_path):
        self.payload['OutputPath'] = output_path
        return self

    def result_path(self, result_path):
        self.payload['ResultPath'] = result_path
        return self


import os

from metaflow.exception import MetaflowException
from metaflow.decorators import StepDecorator


class EnvironmentDecorator(StepDecorator):
    """
    Step decorator to add or update environment variables prior to the execution of your step.

    The environment variables set with this decorator will be present during the execution of the
    step.

    To use, annotate your step as follows:
    ```
    @environment(vars={'MY_ENV': 'value'})
    @step
    def myStep(self):
        ...
    ```

    Parameters
    ----------
    vars : Dict
        Dictionary of environment variables to add/update prior to executing your step.
    """
    name = 'environment'
    defaults = {'vars': {}}

    def runtime_step_cli(self, cli_args, retry_count, max_user_code_retries, ubf_context):
        cli_args.env.update(self.attributes['vars'].items())


import functools
import os
import sys

# HACK to pretend that we installed test_lib
sys.path.append(os.path.realpath(
    os.path.join(os.path.dirname(__file__), '..', 'test_lib_impl')))

import test_lib as lib

EXPORTED_CLASSES = {
    "test_lib": {
        "TestClass1": lib.TestClass1,
        "TestClass2": lib.TestClass2,
        "package.TestClass3": lib.TestClass3,
    }
}

EXPORTED_EXCEPTIONS = {
    "test_lib": {
        "SomeException": lib.SomeException,
        "MyBaseException": lib.MyBaseException,
    }
}

PROXIED_CLASSES = [functools.partial]

EXPORTED_FUNCTIONS = {"test_lib": {"test_func": lib.test_func}}

EXPORTED_VALUES = {"test_lib": {"test_value": lib.test_value}}



import sys
from .mflog import decorate

# This script is similar to the command-line utility 'tee':
# It reads stdin line by line and writes the lines to stdout
# and a file. In contrast to 'tee', this script formats each
# line with mflog-style structure.

if __name__ == '__main__':
    SOURCE = sys.argv[1].encode('ascii')

    with open(sys.argv[2], mode='ab', buffering=0) as f:
        if sys.version_info < (3, 0):
            # Python 2
            for line in iter(sys.stdin.readline, ''):
                # https://bugs.python.org/issue3907
                decorated = decorate(SOURCE, line)
                f.write(decorated)
                sys.stdout.write(line)
        else:
            # Python 3
            for line in sys.stdin.buffer:
                decorated = decorate(SOURCE, line)
                f.write(decorated)
                sys.stdout.buffer.write(line)


from setuptools import setup, find_packages

version = '2.3.4'

setup(name='metaflow',
      version=version,
      description='Metaflow: More Data Science, Less Engineering',
      author='Machine Learning Infrastructure Team at Netflix',
      author_email='help@metaflow.org',
      license='Apache License 2.0',
      packages=find_packages(exclude=['metaflow_test']),
      py_modules=['metaflow', ],
      package_data={'metaflow' : ['tutorials/*/*']},
      entry_points='''
        [console_scripts]
        metaflow=metaflow.main_cli:main
      ''',
      install_requires = [
        'click>=7.0',
        'requests',
        'boto3',
        'pylint'
      ],
      tests_require = [
        'coverage'
      ])



from __future__ import print_function

import json
import time
import math
import sys
import os
import traceback
from hashlib import sha1
from tempfile import NamedTemporaryFile
from multiprocessing import Process, Queue
from itertools import starmap, chain, islice

try:
    # python2
    from urlparse import urlparse
    from Queue import Full as QueueFull
except:
    # python3
    from urllib.parse import urlparse
    from queue import Full as QueueFull

import click

# s3op can be launched as a stand-alone script. We must set
# PYTHONPATH for the parent Metaflow explicitly.
sys.path.insert(0,\
    os.path.abspath(os.path.join(os.path.dirname(__file__), '../../')))
# we use Metaflow's parallel_imap_unordered instead of
# multiprocessing.Pool because https://bugs.python.org/issue31886
from metaflow.util import TempDir, url_quote, url_unquote
from metaflow.multicore_utils import parallel_map
from metaflow.datastore.util.s3util import aws_retry

NUM_WORKERS_DEFAULT = 64

class S3Url(object):
    def __init__(self,
        bucket, path, url, local, prefix,
        content_type=None, metadata=None, range=None):

        self.bucket = bucket
        self.path = path
        self.url = url
        self.local = local
        self.prefix = prefix
        self.content_type = content_type
        self.metadata = metadata
        self.range = range

    def __str__(self):
        return self.url

# We use error codes instead of Exceptions, which are trickier to
# handle reliably in a multi-process world
ERROR_INVALID_URL = 4
ERROR_NOT_FULL_PATH = 5
ERROR_URL_NOT_FOUND = 6
ERROR_URL_ACCESS_DENIED = 7
ERROR_WORKER_EXCEPTION = 8
ERROR_VERIFY_FAILED = 9
ERROR_LOCAL_FILE_NOT_FOUND = 10

def format_triplet(prefix, url='', local=''):
    return u' '.join(url_quote(x).decode('utf-8') for x in (prefix, url, local))

# I can't understand what's the right way to deal
# with boto errors. This function can be replaced
# with better error handling code.
def normalize_client_error(err):
    error_code = err.response['Error']['Code']
    try:
        return int(error_code)
    except ValueError:
        if error_code == 'AccessDenied':
            return 403
    return error_code

# S3 worker pool

def worker(result_file_name, queue, mode):
    # Interpret mode, it can either be a single op or something like
    # info_download or info_upload which implies:
    #  - for download: we need to return the information as well
    #  - for upload: we need to not overwrite the file if it exists
    modes = mode.split('_')
    pre_op_info = False
    if len(modes) > 1:
        pre_op_info = True
        mode = modes[1]
    else:
        mode = modes[0]

    def op_info(url):
        try:
            head = s3.head_object(Bucket=url.bucket, Key=url.path)
            to_return = {
                'error': None,
                'size': head['ContentLength'],
                'content_type': head['ContentType'],
                'metadata': head['Metadata']}
        except client_error as err:
            error_code = normalize_client_error(err)
            if error_code == 404:
                to_return = {'error': ERROR_URL_NOT_FOUND, 'raise_error': err}
            elif error_code == 403:
                to_return = {'error': ERROR_URL_ACCESS_DENIED, 'raise_error': err}
            else:
                to_return = {'error': error_code, 'raise_error': err}
        return to_return

    with open(result_file_name, 'w') as result_file:
        try:
            from metaflow.datastore.util.s3util import get_s3_client
            s3, client_error = get_s3_client()
            while True:
                url, idx = queue.get()
                if url is None:
                    break
                if mode == 'info':
                    result = op_info(url)
                    orig_error = result.get('raise_error', None)
                    if orig_error:
                        del result['raise_error']
                    with open(url.local, 'w') as f:
                        json.dump(result, f)
                elif mode == 'download':
                    result_info = None
                    is_missing = False
                    if pre_op_info:
                        result_info = op_info(url)
                        if result_info['error'] == ERROR_URL_NOT_FOUND:
                            is_missing = True
                            result_file.write("%d %d\n" % (idx, -ERROR_URL_NOT_FOUND))
                        elif result_info['error'] == ERROR_URL_ACCESS_DENIED:
                            is_missing = True
                            result_file.write("%d %d\n" % (idx, -ERROR_URL_ACCESS_DENIED))
                        elif result_info['error'] is not None:
                            raise result_info['raise_error']
                    if is_missing:
                        continue
                    tmp = NamedTemporaryFile(dir='.', delete=False)
                    try:
                        if url.range is None:
                            s3.download_file(url.bucket, url.path, tmp.name)
                        else:
                            # We do get_object. We don't actually do any retries
                            # here because the higher levels will do the retry if
                            # needed
                            resp = s3.get_object(
                                Bucket=url.bucket,
                                Key=url.path,
                                Range=url.range)
                            code = str(resp['ResponseMetadata']['HTTPStatusCode'])
                            if code[0] == '2':
                                tmp.write(resp['Body'].read())
                            else:
                                # TODO: Better raised error
                                raise RuntimeError("Could not load file")
                        tmp.close()
                        os.rename(tmp.name, url.local)
                    except client_error as err:
                        error_code = normalize_client_error(err)
                        if error_code == 404:
                            pass # We skip this
                        else:
                            raise
                    except:
                        # TODO specific error message for out of disk space
                        tmp.close()
                        os.unlink(tmp.name)
                        raise
                    # If we have metadata that we retrieved, we also write it out
                    # to a file
                    if result_info:
                        with open('%s_meta' % url.local, mode='w') as f:
                            args = {'size': result_info['size']}
                            if result_info['content_type']:
                                args['content_type'] = result_info['content_type']
                            if result_info['metadata'] is not None:
                                args['metadata'] = result_info['metadata']
                            json.dump(args, f)
                        # Finally, we push out the size to the result_pipe since
                        # the size is used for verification and other purposes and
                        # we want to avoid file operations for this simple process
                        result_file.write("%d %d\n" % (idx, result_info['size']))
                else:
                    # This is upload, if we have a pre_op, it means we do not
                    # want to overwrite
                    do_upload = False
                    if pre_op_info:
                        result_info = op_info(url)
                        if result_info['error'] == ERROR_URL_NOT_FOUND:
                            # We only upload if the file is not found
                            do_upload = True
                    else:
                        # No pre-op so we upload
                        do_upload = True
                    if do_upload:
                        extra=None
                        if url.content_type or url.metadata:
                            extra = {}
                            if url.content_type:
                                extra['ContentType'] = url.content_type
                            if url.metadata is not None:
                                extra['Metadata'] = url.metadata
                        s3.upload_file(url.local, url.bucket, url.path, ExtraArgs=extra)
                        # We indicate that the file was uploaded
                        result_file.write("%d %d\n" % (idx, 0))
        except:
            traceback.print_exc()
            sys.exit(ERROR_WORKER_EXCEPTION)

def start_workers(mode, urls, num_workers):
    # We start the minimum of len(urls) or num_workers to avoid starting
    # workers that will definitely do nothing
    num_workers = min(num_workers, len(urls))
    queue = Queue(len(urls) + num_workers)
    procs = {}

    # 1. push sources and destinations to the queue
    for idx, elt in enumerate(urls):
        queue.put((elt, idx))

    # 2. push end-of-queue markers
    for i in range(num_workers):
        queue.put((None, None))

    # 3. Prepare the result structure
    sz_results = [None]*len(urls)

    # 4. start processes
    with TempDir() as output_dir:
        for i in range(num_workers):
            file_path = os.path.join(output_dir, str(i))
            p = Process(target=worker, args=(file_path, queue, mode))
            p.start()
            procs[p] = file_path

        # 5. wait for the processes to finish; we continuously update procs
        # to remove all processes that have finished already
        while procs:
            new_procs = {}
            for proc, out_path in procs.items():
                proc.join(timeout=1)
                if proc.exitcode is not None:
                    if proc.exitcode != 0:
                        msg = 'Worker process failed (exit code %d)'\
                                % proc.exitcode
                        exit(msg, proc.exitcode)
                    # Read the output file if all went well
                    with open(out_path, 'r') as out_file:
                        for line in out_file:
                            line_split = line.split(' ')
                            sz_results[int(line_split[0])] = int(line_split[1])
                else:
                    # Put this process back in the processes to check
                    new_procs[proc] = out_path
            procs = new_procs
    return sz_results

def process_urls(mode, urls, verbose, num_workers):

    if verbose:
        print('%sing %d files..' % (mode.capitalize(), len(urls)),
              file=sys.stderr)

    start = time.time()
    sz_results = start_workers(mode, urls, num_workers)
    end = time.time()

    if verbose:
        total_size = sum(sz for sz in sz_results if sz is not None and sz > 0)
        bw = total_size / (end - start)
        print('%sed %d files, %s in total, in %d seconds (%s/s).'\
              % (mode.capitalize(),
                 len(urls),
                 with_unit(total_size),
                 end - start,
                 with_unit(bw)),
              file=sys.stderr)
    return sz_results

# Utility functions

def with_unit(x):
    if x > 1024**3:
        return '%.1fGB' % (x / 1024.**3)
    elif x > 1024**2:
        return '%.1fMB' % (x / 1024.**2)
    elif x > 1024:
        return '%.1fKB' % (x / 1024.)
    else:
        return '%d bytes' % x

# S3Ops class is just a wrapper for get_size and list_prefix
# required by @aws_retry decorator, which needs the reset_client
# method. Otherwise they would be just stand-alone functions.
class S3Ops(object):

    def __init__(self):
        self.s3 = None
        self.client_error = None

    def reset_client(self, hard_reset=False):
        from metaflow.datastore.util.s3util import get_s3_client
        if hard_reset or self.s3 is None:
            self.s3, self.client_error = get_s3_client()

    @aws_retry
    def get_info(self, url):
        self.reset_client()
        try:
            head = self.s3.head_object(Bucket=url.bucket, Key=url.path)
            return True, url, [(S3Url(
                bucket=url.bucket,
                path=url.path,
                url=url.url,
                local=url.local,
                prefix=url.prefix,
                content_type=head['ContentType'],
                metadata=head['Metadata'],
                range=url.range), head['ContentLength'])]
        except self.client_error as err:
            error_code = normalize_client_error(err)
            if error_code == 404:
                return False, url, ERROR_URL_NOT_FOUND
            elif error_code == 403:
                return False, url, ERROR_URL_ACCESS_DENIED
            else:
                raise

    @aws_retry
    def list_prefix(self, prefix_url, delimiter=''):
        self.reset_client()
        url_base = 's3://%s/' % prefix_url.bucket
        try:
            paginator = self.s3.get_paginator('list_objects_v2')
            urls = []
            for page in paginator.paginate(Bucket=prefix_url.bucket,
                                           Prefix=prefix_url.path,
                                           Delimiter=delimiter):
                # note that an url may be both a prefix and an object
                # - the trailing slash is significant in S3
                if 'Contents' in page:
                    for key in page.get('Contents', []):
                        url = url_base + key['Key']
                        urlobj = S3Url(url=url,
                                       bucket=prefix_url.bucket,
                                       path=key['Key'],
                                       local=generate_local_path(url),
                                       prefix=prefix_url.url)
                        urls.append((urlobj, key['Size']))
                if 'CommonPrefixes' in page:
                    # we get CommonPrefixes if Delimiter is a non-empty string
                    for key in page.get('CommonPrefixes', []):
                        url = url_base + key['Prefix']
                        urlobj = S3Url(url=url,
                                       bucket=prefix_url.bucket,
                                       path=key['Prefix'],
                                       local=None,
                                       prefix=prefix_url.url)
                        urls.append((urlobj, None))
            return True, prefix_url, urls
        except self.s3.exceptions.NoSuchBucket:
            return False, prefix_url, ERROR_URL_NOT_FOUND
        except self.client_error as err:
            if err.response['Error']['Code'] == 'AccessDenied':
                return False, prefix_url, ERROR_URL_ACCESS_DENIED
            else:
                raise

# We want to reuse an s3 client instance over multiple operations.
# This is accomplished by op_ functions below.

def op_get_info(urls):
    s3 = S3Ops()
    return [s3.get_info(url) for url in urls]

def op_list_prefix(prefix_urls):
    s3 = S3Ops()
    return [s3.list_prefix(prefix) for prefix in prefix_urls]

def op_list_prefix_nonrecursive(prefix_urls):
    s3 = S3Ops()
    return [s3.list_prefix(prefix, delimiter='/') for prefix in prefix_urls]

def exit(exit_code, url):
    if exit_code == ERROR_INVALID_URL:
        msg = 'Invalid url: %s' % url.url
    elif exit_code == ERROR_NOT_FULL_PATH:
        msg = 'URL not a full path: %s' % url.url
    elif exit_code == ERROR_URL_NOT_FOUND:
        msg = 'URL not found: %s' % url.url
    elif exit_code == ERROR_URL_ACCESS_DENIED:
        msg = 'Access denied to URL: %s' % url.url
    elif exit_code == ERROR_WORKER_EXCEPTION:
        msg = 'Download failed'
    elif exit_code == ERROR_VERIFY_FAILED:
        msg = 'Verification failed for URL %s, local file %s'\
              % (url.url, url.local)
    elif exit_code == ERROR_LOCAL_FILE_NOT_FOUND:
        msg = 'Local file not found: %s' % url
    else:
        msg = 'Unknown error'
    print('s3op failed:\n%s' % msg, file=sys.stderr)
    sys.exit(exit_code)

def verify_results(urls, verbose=False):
    for url, expected in urls:
        if verbose:
            print('verifying %s, expected %s' % (url, expected),
                  file=sys.stderr)
        try:
            got = os.stat(url.local).st_size
        except OSError:
            raise
            exit(ERROR_VERIFY_FAILED, url)
        if expected != got:
            exit(ERROR_VERIFY_FAILED, url)
        if url.content_type or url.metadata:
            # Verify that we also have a metadata file present
            try:
                os.stat('%s_meta' % url.local)
            except OSError:
                exit(ERROR_VERIFY_FAILED, url)

def generate_local_path(url, suffix=None):
    # this function generates a safe local file name corresponding to
    # an S3 URL. URLs may be longer than maximum file length limit on Linux,
    # so we mostly hash the URL but retain the leaf part as a convenience
    # feature to ease eyeballing
    quoted = url_quote(url)
    fname = quoted.split(b'/')[-1].replace(b'.', b'_').replace(b'-', b'_')
    sha = sha1(quoted).hexdigest()
    if suffix:
        return u'-'.join((sha, fname.decode('utf-8'), suffix))
    return u'-'.join((sha, fname.decode('utf-8')))

def parallel_op(op, lst, num_workers):
    # parallel op divides work equally amongst num_workers
    # processes. This is a good strategy if the cost is
    # uniform over the units of work, e.g. op_get_info, which
    # is a single HEAD request to S3.
    #
    # This approach is less optimal with op_list_prefix where
    # the cost of S3 listing per prefix can vary drastically.
    # We could optimize this case by using a worker model with
    # a queue, like for downloads but the difference here is
    # that we need to return a value, which would require a
    # bit more work - something to consider if this turns out
    # to be a bottleneck.
    if lst:
        num = min(len(lst), num_workers)
        batch_size = math.ceil(len(lst) / float(num))
        batches = []
        it = iter(lst)
        while True:
            batch = list(islice(it, batch_size))
            if batch:
                batches.append(batch)
            else:
                break
        it = parallel_map(op, batches, max_parallel=num)
        for x in chain.from_iterable(it):
            yield x

# CLI

@click.group()
def cli():
    pass

@cli.command('list', help='List S3 objects')
@click.option('--inputs',
              type=click.Path(exists=True),
              help='Read input prefixes from the given file.')
@click.option('--num-workers',
              default=NUM_WORKERS_DEFAULT,
              show_default=True,
              help='Number of concurrent connections.')
@click.option('--recursive/--no-recursive',
              default=False,
              show_default=True,
              help='Download prefixes recursively.')
@click.argument('prefixes', nargs=-1)
def lst(prefixes,
        inputs=None,
        num_workers=None,
        recursive=None):

    urllist = []
    for prefix, _ in _populate_prefixes(prefixes, inputs):
        src = urlparse(prefix)
        url = S3Url(url=prefix,
                    bucket=src.netloc,
                    path=src.path.lstrip('/'),
                    local=None,
                    prefix=prefix)
        if src.scheme != 's3':
            exit(ERROR_INVALID_URL, url)
        urllist.append(url)

    op = op_list_prefix if recursive else op_list_prefix_nonrecursive
    urls = []
    for success, prefix_url, ret in parallel_op(op, urllist, num_workers):
        if success:
            urls.extend(ret)
        else:
            exit(ret, prefix_url)

    for url, size in urls:
        if size is None:
            print(format_triplet(url.prefix, url.url))
        else:
            print(format_triplet(url.prefix, url.url, str(size)))

@cli.command(help='Upload files to S3')
@click.option('--file',
              'files',
              type=(click.Path(exists=True), str),
              multiple=True,
              help='Local file->S3Url pair to upload. '
                   'Can be specified multiple times.')
@click.option('--filelist',
              type=click.Path(exists=True),
              help='Read local file -> S3 URL mappings from the given file.')
@click.option('--num-workers',
              default=NUM_WORKERS_DEFAULT,
              show_default=True,
              help='Number of concurrent connections.')
@click.option('--verbose/--no-verbose',
              default=True,
              show_default=True,
              help='Print status information on stderr.')
@click.option('--overwrite/--no-overwrite',
              default=True,
              show_default=True,
              help='Overwrite key if it already exists in S3.')
@click.option('--listing/--no-listing',
              default=False,
              show_default=True,
              help='Print S3 URLs upload to on stdout.')
def put(files=None,
        filelist=None,
        num_workers=None,
        verbose=None,
        overwrite=True,
        listing=None):

    def _files():
        for local, url in files:
            yield url_unquote(local), url_unquote(url), None, None
        if filelist:
            for line in open(filelist, mode='rb'):
                r = json.loads(line)
                local = r['local']
                url = r['url']
                content_type = r.get('content_type', None)
                metadata = r.get('metadata', None)
                if not os.path.exists(local):
                    exit(ERROR_LOCAL_FILE_NOT_FOUND, local)
                yield local, url, content_type, metadata

    def _make_url(local, user_url, content_type, metadata):
        src = urlparse(user_url)
        url = S3Url(url=user_url,
                    bucket=src.netloc,
                    path=src.path.lstrip('/'),
                    local=local,
                    prefix=None,
                    content_type=content_type,
                    metadata=metadata)
        if src.scheme != 's3':
            exit(ERROR_INVALID_URL, url)
        if not src.path:
            exit(ERROR_NOT_FULL_PATH, url)
        return url

    urls = list(starmap(_make_url, _files()))
    ul_op = 'upload'
    if not overwrite:
        ul_op = 'info_upload'
    sz_results = process_urls(ul_op, urls, verbose, num_workers)
    urls = [url for url, sz in zip(urls, sz_results) if sz is not None]
    if listing:
        for url in urls:
            print(format_triplet(url.url))

def _populate_prefixes(prefixes, inputs):
    # Returns a tuple: first element is the prefix and second element
    # is the optional range (or None if the entire prefix is requested)
    if prefixes:
        prefixes = [(url_unquote(p), None) for p in prefixes]
    else:
        prefixes = []
    if inputs:
        with open(inputs, mode='rb') as f:
            for l in f:
                s = l.split(b' ')
                if len(s) > 1:
                    prefixes.append(
                        (url_unquote(s[0].strip()), url_unquote(s[1].strip())))
                else:
                    prefixes.append((url_unquote(s[0].strip()), None))
    return prefixes

@cli.command(help='Download files from S3')
@click.option('--recursive/--no-recursive',
              default=False,
              show_default=True,
              help='Download prefixes recursively.')
@click.option('--num-workers',
              default=NUM_WORKERS_DEFAULT,
              show_default=True,
              help='Number of concurrent connections.')
@click.option('--inputs',
              type=click.Path(exists=True),
              help='Read input prefixes from the given file.')
@click.option('--verify/--no-verify',
              default=True,
              show_default=True,
              help='Verify that files were loaded correctly.')
@click.option('--info/--no-info',
              default=True,
              show_default=True,
              help='Return user tags and content-type')
@click.option('--allow-missing/--no-allow-missing',
              default=False,
              show_default=True,
              help='Do not exit if missing files are detected. '\
                   'Implies --verify.')
@click.option('--verbose/--no-verbose',
              default=True,
              show_default=True,
              help='Print status information on stderr.')
@click.option('--listing/--no-listing',
              default=False,
              show_default=True,
              help='Print S3 URL -> local file mapping on stdout.')
@click.argument('prefixes', nargs=-1)
def get(prefixes,
        recursive=None,
        num_workers=None,
        inputs=None,
        verify=None,
        info=None,
        allow_missing=None,
        verbose=None,
        listing=None):

    # Construct a list of URL (prefix) objects
    urllist = []
    for prefix, r in _populate_prefixes(prefixes, inputs):
        src = urlparse(prefix)
        url = S3Url(url=prefix,
                    bucket=src.netloc,
                    path=src.path.lstrip('/'),
                    local=generate_local_path(prefix),
                    prefix=prefix,
                    range=r)
        if src.scheme != 's3':
            exit(ERROR_INVALID_URL, url)
        if not recursive and not src.path:
            exit(ERROR_NOT_FULL_PATH, url)
        urllist.append(url)
    # Construct a url->size mapping and get content-type and metadata if needed
    op = None
    dl_op = 'download'
    if recursive:
        op = op_list_prefix
    if verify or verbose or info:
        dl_op = 'info_download'
    if op:
        urls = []
        # NOTE - we must retain the order of prefixes requested
        # and the listing order returned by S3
        for success, prefix_url, ret in parallel_op(op, urllist, num_workers):
            if success:
                urls.extend(ret)
            elif ret == ERROR_URL_NOT_FOUND and allow_missing:
                urls.append((prefix_url, None))
            else:
                exit(ret, prefix_url)
    else:
        # pretend zero size since we don't need it for anything.
        # it can't be None though, to make sure the listing below
        # works correctly (None denotes a missing file)
        urls = [(prefix_url, 0) for prefix_url in urllist]

    # exclude the non-existent files from loading
    to_load = [url for url, size in urls if size is not None]
    sz_results = process_urls(dl_op, to_load, verbose, num_workers)
    # We check if there is any access denied
    is_denied = [sz == -ERROR_URL_ACCESS_DENIED for sz in sz_results]
    if any(is_denied):
        # Find the first one to return that as an error
        for i, b in enumerate(is_denied):
            if b:
                exit(ERROR_URL_ACCESS_DENIED, to_load[i])
    if not allow_missing:
        is_missing = [sz == -ERROR_URL_NOT_FOUND for sz in sz_results]
        if any(is_missing):
            # Find the first one to return that as an error
            for i, b in enumerate(is_missing):
                if b:
                    exit(ERROR_URL_NOT_FOUND, to_load[i])
    # Postprocess
    if verify:
        # Verify only results with an actual size (so actual files)
        verify_results([(url, sz) for url, sz in zip(to_load, sz_results)
                        if sz != -ERROR_URL_NOT_FOUND], verbose=verbose)

    idx_in_sz = 0
    if listing:
        for url, _ in urls:
            sz = None
            if idx_in_sz != len(to_load) and url.url == to_load[idx_in_sz].url:
                sz = sz_results[idx_in_sz] if sz_results[idx_in_sz] >= 0 else None
                idx_in_sz += 1
            if sz is None:
                # This means that either the initial url had a None size or
                # that after loading, we found a None size
                print(format_triplet(url.url))
            else:
                print(format_triplet(url.prefix, url.url, url.local))

@cli.command(help='Get info about files from S3')
@click.option('--num-workers',
              default=NUM_WORKERS_DEFAULT,
              show_default=True,
              help='Number of concurrent connections.')
@click.option('--inputs',
              type=click.Path(exists=True),
              help='Read input prefixes from the given file.')
@click.option('--verbose/--no-verbose',
              default=True,
              show_default=True,
              help='Print status information on stderr.')
@click.option('--listing/--no-listing',
              default=False,
              show_default=True,
              help='Print S3 URL -> local file mapping on stdout.')
@click.argument('prefixes', nargs=-1)
def info(prefixes,
         num_workers=None,
         inputs=None,
         verbose=None,
         listing=None):

    # Construct a list of URL (prefix) objects
    urllist = []
    for prefix, _ in _populate_prefixes(prefixes, inputs):
        src = urlparse(prefix)
        url = S3Url(url=prefix,
                    bucket=src.netloc,
                    path=src.path.lstrip('/'),
                    local=generate_local_path(prefix, suffix='info'),
                    prefix=prefix,
                    range=None)
        if src.scheme != 's3':
            exit(ERROR_INVALID_URL, url)
        urllist.append(url)

    process_urls('info', urllist, verbose, num_workers)

    if listing:
        for url in urllist:
            print(format_triplet(url.prefix, url.url, url.local))

if __name__ == '__main__':
    cli(auto_envvar_prefix='S3OP')



from metaflow import FlowSpec, step, IncludeFile


def script_path(filename):
    """
    A convenience function to get the absolute path to a file in this
    tutorial's directory. This allows the tutorial to be launched from any
    directory.

    """
    import os

    filepath = os.path.join(os.path.dirname(__file__))
    return os.path.join(filepath, filename)


class MovieStatsFlow(FlowSpec):
    """
    A flow to generate some statistics about the movie genres.

    The flow performs the following steps:
    1) Ingests a CSV into a Pandas Dataframe.
    2) Fan-out over genre using Metaflow foreach.
    3) Compute quartiles for each genre.
    4) Save a dictionary of genre specific statistics.

    """
    movie_data = IncludeFile("movie_data",
                             help="The path to a movie metadata file.",
                             default=script_path('movies.csv'))

    @step
    def start(self):
        """
        The start step:
        1) Loads the movie metadata into pandas dataframe.
        2) Finds all the unique genres.
        3) Launches parallel statistics computation for each genre.

        """
        import pandas
        from io import StringIO

        # Load the data set into a pandas dataframe.
        self.dataframe = pandas.read_csv(StringIO(self.movie_data))

        # The column 'genres' has a list of genres for each movie. Let's get
        # all the unique genres.
        self.genres = {genre for genres \
                       in self.dataframe['genres'] \
                       for genre in genres.split('|')}
        self.genres = list(self.genres)

        # We want to compute some statistics for each genre. The 'foreach'
        # keyword argument allows us to compute the statistics for each genre in
        # parallel (i.e. a fan-out).
        self.next(self.compute_statistics, foreach='genres')

    @step
    def compute_statistics(self):
        """
        Compute statistics for a single genre.

        """
        # The genre currently being processed is a class property called
        # 'input'.
        self.genre = self.input
        print("Computing statistics for %s" % self.genre)

        # Find all the movies that have this genre and build a dataframe with
        # just those movies and just the columns of interest.
        selector = self.dataframe['genres'].\
                   apply(lambda row: self.genre in row)
        self.dataframe = self.dataframe[selector]
        self.dataframe = self.dataframe[['movie_title', 'genres', 'gross']]

        # Get some statistics on the gross box office for these titles.
        points = [.25, .5, .75]
        self.quartiles = self.dataframe['gross'].quantile(points).values

        # Join the results from other genres.
        self.next(self.join)

    @step
    def join(self, inputs):
        """
        Join our parallel branches and merge results into a dictionary.

        """
        # Merge results from the genre specific computations.
        self.genre_stats = {inp.genre.lower(): \
                            {'quartiles': inp.quartiles,
                             'dataframe': inp.dataframe} \
                            for inp in inputs}

        self.next(self.end)

    @step
    def end(self):
        """
        End the flow.

        """
        pass


if __name__ == '__main__':
    MovieStatsFlow()



import base64
import functools
import pickle
import sys

from collections import OrderedDict, defaultdict, namedtuple
from copy import copy
from datetime import datetime

ObjReference = namedtuple("ObjReference", "value_type class_name identifier")

# This file encodes/decodes values to send over the wire

if sys.version_info[0] >= 3:

    class InvalidLong:
        pass

    class InvalidUnicode:
        pass


# Types that we can marshall/unmarshall
_types = [
    type(None),
    bool,
    int,
    float,
    complex,
    str,
    list,
    tuple,
    bytearray,
    bytes,
    set,
    frozenset,
    dict,
    defaultdict,
    OrderedDict,
    datetime
]

_container_types = (list, tuple, set, frozenset, dict, defaultdict, OrderedDict)

if sys.version_info[0] >= 3:
    _types.extend([InvalidLong, InvalidUnicode])
    _simple_types = (bool, int, float, complex, bytearray, bytes, datetime)
else:
    _types.extend([long, unicode])  # noqa F821
    _simple_types = (
        bool,
        int,
        float,
        complex,
        bytearray,
        bytes,
        unicode,  # noqa F821
        long,  # noqa F821
        datetime
    )

_types_to_encoding = {x: idx for idx, x in enumerate(_types)}

_dumpers = {}  # Key: type -> Value: function to call to dump
_loaders = {}  # Key: int -> Value: function to call to load that type

FIELD_TYPE = "t"
FIELD_ANNOTATION = "a"  # Additional information that loader/dumper can use
FIELD_INLINE_VALUE = "v"
FIELD_INLINE_KEY = "k"

# Protocol to use
defaultProtocol = pickle.HIGHEST_PROTOCOL


def _register_dumper(what):
    def wrapper(func):
        for w in what:
            _dumpers[w] = functools.partial(func, w)
        return func

    return wrapper


def _register_loader(what):
    def wrapper(func):
        for w in what:
            _loaders[_types_to_encoding[w]] = functools.partial(func, w)
        return func

    return wrapper


@_register_dumper((type(None),))
def _dump_none(obj_type, transferer, obj):
    return (None, False)  # Does not matter what we return


@_register_loader((type(None),))
def _load_none(obj_type, transferer, json_annotation, json_obj):
    return None


@_register_dumper(_simple_types)
def _dump_simple(obj_type, transferer, obj):
    return (None, base64.b64encode(pickle.dumps(obj, protocol=defaultProtocol)).decode("utf-8"))


@_register_loader(_simple_types)
def _load_simple(obj_type, transferer, json_annotation, json_obj):
    new_obj = pickle.loads(base64.b64decode(json_obj), encoding="utf-8")
    if not isinstance(new_obj, obj_type):
        raise RuntimeError("Pickle didn't create an object of the proper type")
    return new_obj


@_register_dumper(_container_types)
def _dump_container(obj_type, transferer, obj):
    try:
        new_obj = transferer.pickle_container(obj)
    except RuntimeError as e:
        raise RuntimeError("Cannot dump container %s: %s" % (str(obj), e))
    if new_obj is None:
        return _dump_simple(obj_type, transferer, obj)
    else:
        _, dump = _dump_simple(obj_type, transferer, new_obj)
        return True, dump


@_register_loader(_container_types)
def _load_container(obj_type, transferer, json_annotation, json_obj):
    obj = _load_simple(obj_type, transferer, json_annotation, json_obj)
    if json_annotation:
        # This means we had pickled references
        obj = transferer.unpickle_container(obj)
    return obj


if sys.version_info[0] >= 3:

    @_register_dumper((str,))
    def _dump_str(obj_type, transferer, obj):
        return _dump_simple(obj_type, transferer, obj)

    @_register_loader((str,))
    def _load_str(obj_type, transferer, json_annotation, json_obj):
        return _load_simple(obj_type, transferer, json_annotation, json_obj)

    @_register_dumper((InvalidLong,))
    def _dump_invalidlong(obj_type, transferer, obj):
        return _dump_simple(int, transferer, obj)

    @_register_loader((InvalidLong,))
    def _load_invalidlong(obj_type, transferer, json_annotation, json_obj):
        return _load_simple(int, transferer, json_annotation, json_obj)

    @_register_dumper((InvalidUnicode,))
    def _dump_invalidunicode(obj_type, transferer, obj):
        return _dump_simple(str, transferer, obj)

    @_register_loader((InvalidUnicode,))
    def _load_invalidunicode(obj_type, transferer, json_annotation, json_obj):
        return _load_simple(str, transferer, json_annotation, json_obj)


else:

    @_register_dumper((str,))
    def _dump_str(obj_type, transferer, obj):
        return _dump_simple(obj_type, transferer, obj.encode("utf-8"))

    @_register_loader((str,))
    def _load_str(obj_type, transferer, json_annotation, json_obj):
        # The object is actually bytes
        return (_load_simple(bytes, json_annotation, json_obj)).decode("utf-8")

    @_register_dumper((unicode, long))  # noqa F821
    def _dump_py2_simple(obj_type, transferer, obj):
        return _dump_simple(obj_type, transferer, obj)

    @_register_loader((unicode, long))  # noqa F821
    def _load_py2_simple(obj_type, transferer, json_annotation, json_obj):
        return _load_simple(obj_type, transferer, json_annotation, json_obj)


class DataTransferer(object):
    def __init__(self, connection):
        self._dumpers = _dumpers.copy()
        self._loaders = _loaders.copy()
        self._types_to_encoding = _types_to_encoding.copy()

        self._connection = connection

    @staticmethod
    def can_simple_dump(obj):
        return DataTransferer._can_dump(DataTransferer.can_simple_dump, obj)

    def can_dump(self, obj):
        r = DataTransferer._can_dump(self.can_dump, obj)
        if not r:
            return self._connection.can_encode(obj)
        return False

    def dump(self, obj):
        obj_type = type(obj)
        handler = self._dumpers.get(type(obj))
        if handler:
            attr, v = handler(self, obj)
            return {
                FIELD_TYPE: self._types_to_encoding[obj_type],
                FIELD_ANNOTATION: attr,
                FIELD_INLINE_VALUE: v,
            }
        else:
            # We will see if the connection can encode and transfer this object
            # This is primarily used to transfer a reference to an object
            try:
                json_obj = base64.b64encode(
                    pickle.dumps(self._connection.pickle_object(obj),
                                 protocol=defaultProtocol)
                ).decode("utf-8")
            except ValueError as e:
                raise RuntimeError("Unable to dump non base type: %s" % e)
            return {FIELD_TYPE: -1, FIELD_INLINE_VALUE: json_obj}

    def load(self, json_obj):
        obj_type = json_obj.get(FIELD_TYPE)
        if obj_type is None:
            raise RuntimeError(
                "Malformed message -- missing %s: %s" % (FIELD_TYPE, str(json_obj))
            )
        if obj_type == -1:
            # This is something that the connection handles
            try:
                return self._connection.unpickle_object(
                    pickle.loads(base64.b64decode(json_obj[FIELD_INLINE_VALUE]),
                                 encoding="utf-8")
                )
            except ValueError as e:
                raise RuntimeError("Unable to load non base type: %s" % e)
        handler = self._loaders.get(obj_type)
        if handler:
            json_subobj = json_obj.get(FIELD_INLINE_VALUE)
            if json_subobj is not None:
                return handler(
                    self, json_obj.get(FIELD_ANNOTATION), json_obj[FIELD_INLINE_VALUE]
                )
            raise RuntimeError("Non inline value not supported")
        raise RuntimeError("Unable to find handler for type %s" % obj_type)

    # _container_types = (list, tuple, set, frozenset, dict, OrderedDict)
    def _transform_container(self, checker, processor, recursor, obj, in_place=True):
        def _sub_process(obj):
            obj_type = type(obj)
            if obj is None or obj_type in _simple_types or obj_type == str:
                return None
            elif obj_type in _container_types:
                return recursor(obj)
            elif checker(obj):
                return processor(obj)
            else:
                raise RuntimeError(
                    "Cannot pickle object of type %s: %s" % (obj_type, str(obj))
                )

        cast_to = None
        key_change_allowed = True
        update_default_factory = False
        has_changes = False
        if isinstance(obj, (tuple, set, frozenset)):
            cast_to = type(obj)
            obj = list(obj)
            in_place = True # We can do in place since we copied the object
        if isinstance(obj, OrderedDict):
            key_change_allowed = False
        if isinstance(obj, defaultdict):
            # In this case, we use a hack to store the default_factory
            # function inside the dictionary which will get it pickled by
            # the server (as a reference). This is because if this is a lambda
            # it won't pickle well.
            if callable(obj.default_factory):
                if not in_place:
                    obj = copy(obj)
                    in_place = True
                obj['__default_factory'] = obj.default_factory
                obj.default_factory = None
            elif obj.get('__default_factory') is not None:
                # This is in the unpickle path, we need to reset the factory properly
                update_default_factory = True
            has_changes = True
        # We now deal with list or dict
        if isinstance(obj, list):
            for idx in range(len(obj)):
                sub_obj = _sub_process(obj[idx])
                if sub_obj is not None:
                    has_changes = True
                    if not in_place:
                        obj = list(obj)
                        in_place = True
                    obj[idx] = sub_obj
        elif isinstance(obj, dict):
            new_items = {}
            del_keys = []
            for k, v in obj.items():
                sub_key = _sub_process(k)
                if sub_key is not None:
                    if not key_change_allowed:
                        raise RuntimeError(
                            "OrderedDict key cannot contain references -- "
                            "this would change the order"
                        )
                    has_changes = True
                sub_val = _sub_process(v)
                if sub_val is not None:
                    has_changes = True
                if has_changes and not in_place:
                    obj = copy(obj)
                    in_place = True
                if sub_key:
                    if sub_val:
                        new_items[sub_key] = sub_val
                    else:
                        new_items[sub_key] = v
                    del_keys.append(k)
                else:
                    if sub_val:
                        obj[k] = sub_val
            for k in del_keys:
                del obj[k]
            obj.update(new_items)
        else:
            raise RuntimeError("Unknown container type: %s" % type(obj))
        if update_default_factory:
            # We do this here because we now unpickled the reference
            # to default_dict and can set it back up again.
            obj.default_factory = obj['__default_factory']
            del obj['__default_factory']
        if has_changes:
            if cast_to:
                return cast_to(obj)
            return obj
        return None

    def pickle_container(self, obj):
        return self._transform_container(
            self._connection.can_pickle,
            self._connection.pickle_object,
            self.pickle_container,
            obj,
            in_place=False,
        )

    def unpickle_container(self, obj):
        return self._transform_container(
            lambda x: isinstance(x, ObjReference),
            self._connection.unpickle_object,
            self.unpickle_container,
            obj,
        )

    @staticmethod
    def _can_dump(recursive_func, obj):
        obj_type = type(obj)
        if obj is None:
            return True
        if obj_type in _simple_types:
            return True
        if obj_type == str:
            return True
        if obj_type == dict or obj_type == OrderedDict:
            return all(
                (recursive_func(k) and recursive_func(v) for k, v in obj.items())
            )
        if obj_type in _container_types:
            return all((recursive_func(x) for x in obj))
        return False



from metaflow import FlowSpec, step


class HelloFlow(FlowSpec):
    """
    A flow where Metaflow prints 'Hi'.

    Run this flow to validate that Metaflow is installed correctly.

    """
    @step
    def start(self):
        """
        This is the 'start' step. All flows must have a step named 'start' that
        is the first step in the flow.

        """
        print("HelloFlow is starting.")
        self.next(self.hello)

    @step
    def hello(self):
        """
        A step for metaflow to introduce itself.

        """
        print("Metaflow says: Hi!")
        self.next(self.end)

    @step
    def end(self):
        """
        This is the 'end' step. All flows must have an 'end' step, which is the
        last step in the flow.

        """
        print("HelloFlow is all done.")


if __name__ == '__main__':
    HelloFlow()



from metaflow.decorators import StepDecorator
from metaflow.exception import MetaflowException
from metaflow.metaflow_config import MAX_ATTEMPTS


class RetryDecorator(StepDecorator):
    """
    Step decorator to specify that a step should be retried on failure.

    This decorator indicates that if this step fails, it should be retried a certain number of times.

    This decorator is useful if transient errors (like networking issues) are likely in your step.

    This can be used in conjunction with the @retry decorator. In that case, catch will only
    activate if all retries fail and will catch the last exception thrown by the last retry.

    To use, annotate your step as follows:
    ```
    @retry(times=3)
    @step
    def myStep(self):
        ...
    ```

    Parameters
    ----------
    times : int
        Number of times to retry this step. Defaults to 3
    minutes_between_retries : int
        Number of minutes between retries
    """
    name = 'retry'
    defaults = {'times': '3',
                'minutes_between_retries': '2'}

    def step_init(self, flow, graph, step, decos, environment, datastore, logger):
        # The total number of attempts must not exceed MAX_ATTEMPTS.
        # attempts = normal task (1) + retries (N) + @catch fallback (1)
        if int(self.attributes['times']) + 2 > MAX_ATTEMPTS:
            raise MetaflowException('The maximum number of retries is '
                                    '@retry(times=%d).' % (MAX_ATTEMPTS - 2))

    def step_task_retry_count(self):
        return int(self.attributes['times']), 0




from . import local, s3
from .datastore import Inputs, DataException, MetaflowDataStore
from .datastore_set import MetaflowDatastoreSet

DATASTORES = {'local': local.LocalDataStore,
              's3': s3.S3DataStore}



from itertools import islice
import os
import sys
import inspect
import traceback

from . import cmd_with_io
from .parameters import Parameter
from .exception import MetaflowException, MetaflowInternalError, \
    MissingInMergeArtifactsException, UnhandledInMergeArtifactsException
from .graph import FlowGraph
from .unbounded_foreach import UnboundedForeachInput

# For Python 3 compatibility
try:
    basestring
except NameError:
    basestring = str


class InvalidNextException(MetaflowException):
    headline = "Invalid self.next() transition detected"

    def __init__(self, msg):
        # NOTE this assume that InvalidNextException is only raised
        # at the top level of next()
        _, line_no, _, _ = traceback.extract_stack()[-3]
        super(InvalidNextException, self).__init__(msg, line_no)


class FlowSpec(object):
    """
    Main class from which all Flows should inherit.

    Attributes
    ----------
    script_name
    index
    input
    """

    # Attributes that are not saved in the datastore when checkpointing.
    # Name starting with '__', methods, functions and Parameters do not need
    # to be listed.
    _EPHEMERAL = {'_EPHEMERAL',
                  '_datastore',
                  '_cached_input',
                  '_graph',
                  '_flow_decorators',
                  '_steps',
                  'index',
                  'input'}

    _flow_decorators = {}

    def __init__(self, use_cli=True):
        """
        Construct a FlowSpec

        Parameters
        ----------
        use_cli : bool, optional, default: True
            Set to True if the flow is invoked from __main__ or the command line
        """

        self.name = self.__class__.__name__

        self._datastore = None
        self._transition = None
        self._cached_input = {}

        self._graph = FlowGraph(self.__class__)
        self._steps = [getattr(self, node.name) for node in self._graph]

        if use_cli:
            # we import cli here to make sure custom parameters in
            # args.py get fully evaluated before cli.py is imported.
            from . import cli
            cli.main(self)

    @property
    def script_name(self):
        """
        Returns the name of the script containing the flow

        Returns
        -------
        str
            A string containing the name of the script
        """
        fname = inspect.getfile(self.__class__)
        if fname.endswith('.pyc'):
            fname = fname[:-1]
        return os.path.basename(fname)

    def _get_parameters(self):
        for var in dir(self):
            if var[0] == '_':
                continue
            try:
                val = getattr(self, var)
            except:
                continue
            if isinstance(val, Parameter):
                yield var, val

    def _set_datastore(self, datastore):
        self._datastore = datastore

    def __iter__(self):
        """
        Iterate over all steps in the Flow

        Returns
        -------
        Iterator[graph.DAGNode]
            Iterator over the steps in the flow
        """
        return iter(self._steps)

    def __getattr__(self, name):
        if self._datastore and name in self._datastore:
            # load the attribute from the datastore...
            x = self._datastore[name]
            # ...and cache it in the object for faster access
            setattr(self, name, x)
            return x
        else:
            raise AttributeError("Flow %s has no attribute '%s'" %
                                 (self.name, name))

    def cmd(self, cmdline, input={}, output=[]):
        return cmd_with_io.cmd(cmdline,
                               input=input,
                               output=output)

    @property
    def index(self):
        """
        Index of the task in a foreach step

        In a foreach step, multiple instances of this step (tasks) will be executed,
        one for each element in the foreach.
        This property returns the zero based index of the current task. If this is not
        a foreach step, this returns None.

        See Also
        --------
        foreach_stack: A detailed example is given in the documentation of this function

        Returns
        -------
        int
            Index of the task in a foreach step
        """
        if self._foreach_stack:
            return self._foreach_stack[-1].index

    @property
    def input(self):
        """
        Value passed to the task in a foreach step

        In a foreach step, multiple instances of this step (tasks) will be executed,
        one for each element in the foreach.
        This property returns the element passed to the current task. If this is not
        a foreach step, this returns None.

        See Also
        --------
        foreach_stack: A detailed example is given in the documentation of this function

        Returns
        -------
        object
            Input passed to the task (can be any object)
        """
        return self._find_input()

    def foreach_stack(self):
        """
        Returns the current stack of foreach steps for the current step

        This effectively corresponds to the indexes and values at the various levels of nesting.
        For example, considering the following code:
        ```
        @step
        def root(self):
            self.split_1 = ['a', 'b', 'c']
            self.next(self.nest_1, foreach='split_1')

        @step
        def nest_1(self):
            self.split_2 = ['d', 'e', 'f', 'g']
            self.next(self.nest_2, foreach='split_2'):

        @step
        def nest_2(self):
            foo = self.foreach_stack()
        ```
        foo will take the following values in the various tasks for nest_2:
            [(0, 3, 'a'), (0, 4, 'd')]
            [(0, 3, 'a'), (1, 4, 'e')]
            ...
            [(0, 3, 'a'), (3, 4, 'g')]
            [(1, 3, 'b'), (0, 4, 'd')]
            ...

        where each tuple corresponds to:
            - the index of the task for that level of the loop
            - the number of splits for that level of the loop
            - the value for that level of the loop
        Note that the last tuple returned in a task corresponds to:
            - first element: value returned by self.index
            - third element: value returned by self.input

        Returns
        -------
        List[Tuple[int, int, object]]
            An array describing the current stack of foreach steps
        """
        return [(frame.index, frame.num_splits, self._find_input(stack_index=i))
                for i, frame in enumerate(self._foreach_stack)]

    def _find_input(self, stack_index=None):

        if stack_index is None:
            stack_index = len(self._foreach_stack) - 1

        if stack_index in self._cached_input:
            return self._cached_input[stack_index]
        elif self._foreach_stack:
            # NOTE this is obviously an O(n) operation which also requires
            # downloading the whole input data object in order to find the
            # right split. One can override this method with a more efficient
            # input data handler if this is a problem.
            frame = self._foreach_stack[stack_index]
            try:
                var = getattr(self, frame.var)
            except AttributeError:
                # this is where AttributeError happens:
                # [ foreach x ]
                #   [ foreach y ]
                #     [ inner ]
                #   [ join y ] <- call self.foreach_stack here,
                #                 self.x is not available
                self._cached_input[stack_index] = None
            else:
                try:
                    self._cached_input[stack_index] = var[frame.index]
                except TypeError:
                    # __getitem__ not supported, fall back to an iterator
                    self._cached_input[stack_index] = next(islice(var,
                                                                  frame.index,
                                                                  frame.index + 1))
            return self._cached_input[stack_index]

    def merge_artifacts(self, inputs, exclude=[], include=[]):
        """
        Merge the artifacts coming from each merge branch (from inputs)

        This function takes all the artifacts coming from the branches of a
        join point and assigns them to self in the calling step. Only artifacts
        not set in the current step are considered. If, for a given artifact, different
        values are present on the incoming edges, an error will be thrown (and the artifacts
        that "conflict" will be reported).

        As a few examples, in the simple graph: A splitting into B and C and joining in D:
        A:
          self.x = 5
          self.y = 6
        B:
          self.b_var = 1
          self.x = from_b
        C:
          self.x = from_c

        D:
          merge_artifacts(inputs)

        In D, the following artifacts are set:
          - y (value: 6), b_var (value: 1)
          - if from_b and from_c are the same, x will be accessible and have value from_b
          - if from_b and from_c are different, an error will be thrown. To prevent this error,
            you need to manually set self.x in D to a merged value (for example the max) prior to
            calling merge_artifacts.

        Parameters
        ----------
        inputs : List[Steps]
            Incoming steps to the join point
        exclude : List[str], optional
            If specified, do not consider merging artifacts with a name in `exclude`.
            Cannot specify if `include` is also specified
        include : List[str], optional
            If specified, only merge artifacts specified. Cannot specify if `exclude` is
            also specified

        Raises
        ------
        MetaflowException
            This exception is thrown if this is not called in a join step
        UnhandledInMergeArtifactsException
            This exception is thrown in case of unresolved conflicts
        MissingInMergeArtifactsException
            This exception is thrown in case an artifact specified in `include cannot
            be found
        """
        node = self._graph[self._current_step]
        if node.type != 'join':
            msg = "merge_artifacts can only be called in a join and step *{step}* "\
                  "is not a join".format(step=self._current_step)
            raise MetaflowException(msg)
        if len(exclude) > 0 and len(include) > 0:
            msg = "`exclude` and `include` are mutually exclusive in merge_artifacts"
            raise MetaflowException(msg)

        to_merge = {}
        unresolved = []
        for inp in inputs:
            # available_vars is the list of variables from inp that should be considered
            if include:
                available_vars = ((var, sha) for var, sha in inp._datastore.items()
                                  if (var in include) and (not hasattr(self, var)))
            else:
                available_vars = ((var, sha) for var, sha in inp._datastore.items()
                                if (var not in exclude) and (not hasattr(self, var)))
            for var, sha in available_vars:
                _, previous_sha = to_merge.setdefault(var, (inp, sha))
                if previous_sha != sha:
                    # We have a conflict here
                    unresolved.append(var)
        # Check if everything in include is present in to_merge
        missing = []
        for v in include:
            if v not in to_merge and not hasattr(self, v):
                missing.append(v)
        if unresolved:
            # We have unresolved conflicts so we do not set anything and error out
            msg = "Step *{step}* cannot merge the following artifacts due to them "\
                  "having conflicting values:\n[{artifacts}].\nTo remedy this issue, "\
                  "be sure to explictly set those artifacts (using "\
                  "self.<artifact_name> = ...) prior to calling merge_artifacts."\
                  .format(step=self._current_step, artifacts=', '.join(unresolved))
            raise UnhandledInMergeArtifactsException(msg, unresolved)
        if missing:
            msg = "Step *{step}* specifies that [{include}] should be merged but "\
                  "[{missing}] are not present.\nTo remedy this issue, make sure "\
                  "that the values specified in only come from at least one branch"\
                  .format(
                      step=self._current_step,
                      include=', '.join(include),
                      missing=', '.join(missing))
            raise MissingInMergeArtifactsException(msg, missing)
        # If things are resolved, we pass down the variables from the input datastores
        for var, (inp, _) in to_merge.items():
            self._datastore.passdown_partial(inp._datastore, [var])

    def _validate_ubf_step(self, step_name):
        join_list = self._graph[step_name].out_funcs
        if len(join_list) != 1:
            msg = "UnboundedForeach is supported only over a single node, "\
                  "not an arbitrary DAG. Specify a single `join` node"\
                  " instead of multiple:{join_list}."\
                  .format(join_list=join_list)
            raise InvalidNextException(msg)
        join_step = join_list[0]
        join_node = self._graph[join_step]
        join_type = join_node.type

        if join_type != 'join':
            msg = "UnboundedForeach found for:{node} -> {join}."\
                  " The join type isn't valid."\
                  .format(node=step_name,
                          join=join_step)
            raise InvalidNextException(msg)

    def next(self, *dsts, **kwargs):
        """
        Indicates the next step to execute at the end of this step

        This statement should appear once and only once in each and every step (except the `end`
        step). Furthermore, it should be the last statement in the step.

        There are several valid formats to specify the next step:
            - Straight-line connection: self.next(self.next_step) where `next_step` is a method in
              the current class decorated with the `@step` decorator
            - Static fan-out connection: self.next(self.step1, self.step2, ...) where `stepX` are
              methods in the current class decorated with the `@step` decorator
            - Foreach branch:
                self.next(self.foreach_step, foreach='foreach_iterator')
              In this situation, `foreach_step` is a method in the current class decorated with the
              `@step` docorator and `foreach_iterator` is a variable name in the current class that
              evaluates to an iterator. A task will be launched for each value in the iterator and
              each task will execute the code specified by the step `foreach_step`.

        Raises
        ------
        InvalidNextException
            Raised if the format of the arguments does not match one of the ones given above.
        """

        step = self._current_step

        foreach = kwargs.pop('foreach', None)
        condition = kwargs.pop('condition', None)
        if kwargs:
            kw = next(iter(kwargs))
            msg = "Step *{step}* passes an unknown keyword argument "\
                  "'{invalid}' to self.next().".format(step=step, invalid=kw)
            raise InvalidNextException(msg)

        # check: next() is called only once
        if self._transition is not None:
            msg = "Multiple self.next() calls detected in step *{step}*. "\
                  "Call self.next() only once.".format(step=step)
            raise InvalidNextException(msg)

        # check: all destinations are methods of this object
        funcs = []
        for i, dst in enumerate(dsts):
            try:
                name = dst.__func__.__name__
            except:
                msg = "In step *{step}* the {arg}. argument in self.next() is "\
                      "not a function. Make sure all arguments in self.next() "\
                      "are methods of the Flow class."\
                      .format(step=step, arg=i + 1)
                raise InvalidNextException(msg)
            if not hasattr(self, name):
                msg = "Step *{step}* specifies a self.next() transition to an "\
                      "unknown step, *{name}*.".format(step=step,
                                                       name=name)
                raise InvalidNextException(msg)
            funcs.append(name)

        # check: foreach and condition are mutually exclusive
        if not (foreach is None or condition is None):
            msg = "Step *{step}* has an invalid self.next() transition. "\
                  "Specify either 'foreach' or 'condition', not both."\
                  .format(step=step)
            raise InvalidNextException(msg)

        # check: foreach is valid
        if foreach:
            if not isinstance(foreach, basestring):
                msg = "Step *{step}* has an invalid self.next() transition. "\
                      "The argument to 'foreach' must be a string."\
                      .format(step=step)
                raise InvalidNextException(msg)

            if len(dsts) != 1:
                msg = "Step *{step}* has an invalid self.next() transition. "\
                      "Specify exactly one target for 'foreach'."\
                      .format(step=step)
                raise InvalidNextException(msg)

            try:
                foreach_iter = getattr(self, foreach)
            except:
                msg = "Foreach variable *self.{var}* in step *{step}* "\
                      "does not exist. Check your variable."\
                      .format(step=step, var=foreach)
                raise InvalidNextException(msg)

            if issubclass(type(foreach_iter), UnboundedForeachInput):
                self._unbounded_foreach = True
                self._foreach_num_splits = None
                self._validate_ubf_step(funcs[0])
            else:
                try:
                    self._foreach_num_splits = sum(1 for _ in foreach_iter)
                except TypeError:
                    msg = "Foreach variable *self.{var}* in step *{step}* "\
                          "is not iterable. Check your variable."\
                          .format(step=step, var=foreach)
                    raise InvalidNextException(msg)

                if self._foreach_num_splits == 0:
                    msg = "Foreach iterator over *{var}* in step *{step}* "\
                          "produced zero splits. Check your variable."\
                          .format(step=step, var=foreach)
                    raise InvalidNextException(msg)

            self._foreach_var = foreach

        # check: condition is valid
        if condition:
            if not isinstance(condition, basestring):
                msg = "Step *{step}* has an invalid self.next() transition. "\
                      "The argument to 'condition' must be a string."\
                      .format(step=step)
                raise InvalidNextException(msg)
            if len(dsts) != 2:
                msg = "Step *{step}* has an invalid self.next() transition. "\
                      "Specify two targets for 'condition': The first target "\
                      "is used if the condition evaluates to true, the second "\
                      "otherwise.".format(step=step)
                raise InvalidNextException(msg)

        # check: non-keyword transitions are valid
        if foreach is None and condition is None:
            if len(dsts) < 1:
                msg = "Step *{step}* has an invalid self.next() transition. "\
                      "Specify at least one step function as an argument in "\
                      "self.next().".format(step=step)
                raise InvalidNextException(msg)

        self._transition = (funcs, foreach, condition)

    def __str__(self):
        step_name = getattr(self, '_current_step', None)
        if step_name:
            index = ','.join(str(idx) for idx, _, _ in self.foreach_stack())
            if index:
                inp = self.input
                if inp is None:
                    return '<flow %s step %s[%s]>' %\
                           (self.name, step_name, index)
                else:
                    inp = str(inp)
                    if len(inp) > 20:
                        inp = inp[:20] + '...'
                    return '<flow %s step %s[%s] (input: %s)>' %\
                        (self.name, step_name, index, inp)
            else:
                return '<flow %s step %s>' % (self.name, step_name)
        else:
            return '<flow %s>' % self.name

    def __getstate__(self):
        raise MetaflowException("Flows can't be serialized. Maybe you tried "
                                "to assign *self* or one of the *inputs* "
                                "to an attribute? Instead of serializing the "
                                "whole flow, you should choose specific "
                                "attributes, e.g. *input.some_var*, to be "
                                "stored.")



from metaflow import FlowSpec, step, Parameter


class PlayListFlow(FlowSpec):
    """
    The next version of our playlist generator that uses the statistics
    generated from 'Episode 02' to improve the title recommendations.

    The flow performs the following steps:

    1) Load the genre specific statistics from the MovieStatsFlow.
    2) In parallel branches:
       - A) Build a playlist from the top grossing films in the requested genre.
       - B) Choose a random movie.
    3) Join the two to create a movie playlist and display it.

    """
    genre = Parameter('genre',
                      help="Filter movies for a particular genre.",
                      default='Sci-Fi')

    recommendations = Parameter('recommendations',
                                help="The number of movies recommended for "
                                "the playlist.",
                                default=5)

    @step
    def start(self):
        """
        Use the Metaflow client to retrieve the latest successful run from our
        MovieStatsFlow and assign them as data artifacts in this flow.

        """
        from metaflow import Flow, get_metadata

        # Print metadata provider
        print("Using metadata provider: %s" % get_metadata())

        # Load the analysis from the MovieStatsFlow.
        run = Flow('MovieStatsFlow').latest_successful_run
        print("Using analysis from '%s'" % str(run))

        self.genre_stats = run.data.genre_stats

        # Compute our two recommendation types in parallel.
        self.next(self.bonus_movie, self.genre_movies)

    @step
    def bonus_movie(self):
        """
        This step chooses a random title for a different movie genre.

        """
        import pandas

        # Concatenate all the genre specific data frames and choose a random
        # movie.
        df = pandas.concat([data['dataframe'] \
                            for genre, data in self.genre_stats.items() \
                            if genre != self.genre.lower()])
        df = df.sample(n=1)
        self.bonus = (df['movie_title'].values[0],
                      df['genres'].values[0])

        self.next(self.join)

    @step
    def genre_movies(self):
        """
        Select the top performing movies from the user specified genre.

        """
        from random import shuffle

        # For the genre of interest, generate a potential playlist using only
        # highest gross box office titles (i.e. those in the last quartile).
        genre = self.genre.lower()
        if genre not in self.genre_stats:
            self.movies = []

        else:
            df = self.genre_stats[genre]['dataframe']
            quartiles = self.genre_stats[genre]['quartiles']
            selector = df['gross'] >= quartiles[-1]
            self.movies = list(df[selector]['movie_title'])

        # Shuffle the playlist.
        shuffle(self.movies)

        self.next(self.join)

    @step
    def join(self, inputs):
        """
        Join our parallel branches and merge results.

        """
        self.playlist = inputs.genre_movies.movies
        self.bonus = inputs.bonus_movie.bonus

        self.next(self.end)

    @step
    def end(self):
        """
        Print out the playlist and bonus movie.

        """
        # Print the playlist.
        print("Playlist for movies in genre '%s'" % self.genre)
        for pick, movie in enumerate(self.playlist, start=1):
            print("Pick %d: '%s'" % (pick, movie))
            if pick >= self.recommendations:
                break

        print("Bonus Pick: '%s' from '%s'" % (self.bonus[0], self.bonus[1]))


if __name__ == '__main__':
    PlayListFlow()
